# Copyright 2012,2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import constants as db_const
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin import config
from neutron_tempest_plugin.vpnaas.api import base_vpnaas as base

CONF = config.CONF

_LONG_NAME = 'x' * (db_const.NAME_FIELD_SIZE + 1)
_LONG_DESCRIPTION = 'y' * (db_const.DESCRIPTION_FIELD_SIZE + 1)


class VPNaaSTestJSON(base.BaseAdminNetworkTest):

    """VPNaaS API tests.

    Tests the following operations in the Neutron API using the REST client for
    Neutron:

        * List, Show, Create, Delete, and Update VPN Service
        * List, Show, Create, Delete, and Update IKE policy
        * List, Show, Create, Delete, and Update IPSec policy
    """

    @classmethod
    def resource_setup(cls):
        if not utils.is_extension_enabled('vpnaas', 'network'):
            msg = "vpnaas extension not enabled."
            raise cls.skipException(msg)
        super().resource_setup()
        cls.ext_net_id = CONF.network.public_network_id
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router(
            data_utils.rand_name("router"),
            external_network_id=CONF.network.public_network_id)
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.vpnservice = cls.create_vpnservice(cls.subnet['id'],
                                               cls.router['id'])
        vpnservice2 = cls.create_vpnservice_no_subnet(cls.router['id'])
        cls.vpnservice_no_subnet = vpnservice2

        cls.ikepolicy = cls.create_ikepolicy(
            data_utils.rand_name("ike-policy-"))
        cls.ipsecpolicy = cls.create_ipsecpolicy(
            data_utils.rand_name("ipsec-policy-"))

        cls.endpoint_group_local = cls.create_endpoint_group(
            data_utils.rand_name("endpoint-group-local-"),
            'subnet',
            cls.subnet['id'])

        cls.endpoint_group_remote = cls.create_endpoint_group(
            data_utils.rand_name("endpoint-group-remote-"),
            'cidr',
            ["10.101.0.0/24", "10.102.0.0/24"])

        cls.ipsec_site_connection = cls.create_ipsec_site_connection(
            cls.ikepolicy['id'],
            cls.ipsecpolicy['id'],
            cls.vpnservice['id'])

    def _delete_ike_policy(self, ike_policy_id):
        # Deletes a ike policy and verifies if it is deleted or not
        ike_list = list()
        all_ike = self.client.list_ikepolicies()
        for ike in all_ike['ikepolicies']:
            ike_list.append(ike['id'])
        if ike_policy_id in ike_list:
            self.client.delete_ikepolicy(ike_policy_id)
            # Asserting that the policy is not found in list after deletion
            ikepolicies = self.client.list_ikepolicies()
            ike_id_list = list()
            for i in ikepolicies['ikepolicies']:
                ike_id_list.append(i['id'])
            self.assertNotIn(ike_policy_id, ike_id_list)

    def _delete_ipsec_policy(self, ipsec_policy_id):
        # Deletes an ike policy if it exists
        try:
            self.client.delete_ipsecpolicy(ipsec_policy_id)

        except lib_exc.NotFound:
            pass

    def _delete_ipsec_site_connection(self, conn_id):
        # Deletes an ipsec site connection if it exists
        try:
            self.client.delete_ipsec_site_connection(conn_id)
        except lib_exc.NotFound:
            pass

    def _assertExpected(self, expected, actual):
        # Check if not expected keys/values exists in actual response body
        for key, value in expected.items():
            self.assertIn(key, actual)
            self.assertEqual(value, actual[key])

    def _delete_vpn_service(self, vpn_service_id):
        self.client.delete_vpnservice(vpn_service_id)
        # Asserting if vpn service is found in the list after deletion
        body = self.client.list_vpnservices()
        vpn_services = [vs['id'] for vs in body['vpnservices']]
        self.assertNotIn(vpn_service_id, vpn_services)

    def _delete_endpoint_group(self, endpoint_group_id):
        # Delete a endpoint-group and verifies if it is deleted or not
        endpoint_group_list = list()
        all_endpoint = self.client.list_endpoint_groups()
        for endpoint in all_endpoint['endpoint_groups']:
            endpoint_group_list.append(endpoint['id'])
        if endpoint_group_id in endpoint_group_list:
            self.client.delete_endpoint_group(endpoint_group_id)
            # Asserting that the endpoint is not found in list after deletion
            endpoint_group = self.client.list_endpoint_groups()
            for e in endpoint_group['endpoint_groups']:
                endpoint_group_list.append(e['id'])
            self.assertNotIn(endpoint_group_list, endpoint_group_id)

    def _get_tenant_id(self):
        """Returns the tenant_id of the client current user"""
        return self.client.project_id

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('74dcf2d3-a40e-4a6c-a25a-747d764bee81')
    def test_admin_create_ipsec_policy_for_tenant(self):
        tenant_id = self._get_tenant_id()
        # Create IPSec policy for the newly created tenant
        name = data_utils.rand_name('ipsec-policy')
        body = self.admin_client.create_ipsecpolicy(
            name=name,
            tenant_id=tenant_id,
            auth_algorithm='sha256',
            encryption_algorithm='aes-256',
        )
        ipsecpolicy = body['ipsecpolicy']
        self.assertIsNotNone(ipsecpolicy['id'])
        self.addCleanup(self.admin_client.delete_ipsecpolicy,
                        ipsecpolicy['id'])

        # Assert that created ipsec policy is found in API list call
        body = self.client.list_ipsecpolicies()
        ipsecpolicies = [policy['id'] for policy in body['ipsecpolicies']]
        self.assertIn(ipsecpolicy['id'], ipsecpolicies)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('016b1861-fe55-4184-ba3c-e049ebbeb570')
    def test_admin_create_vpn_service_for_tenant(self):
        tenant_id = self._get_tenant_id()

        # Create vpn service for the newly created tenant
        network2 = self.create_network()
        subnet2 = self.create_subnet(network2)
        router2 = self.create_router(data_utils.rand_name('router-'),
                                     external_network_id=self.ext_net_id)
        self.create_router_interface(router2['id'], subnet2['id'])
        name = data_utils.rand_name('vpn-service')
        body = self.admin_client.create_vpnservice(
            subnet_id=subnet2['id'],
            router_id=router2['id'],
            name=name,
            admin_state_up=True,
            tenant_id=tenant_id)
        vpnservice = body['vpnservice']
        self.assertIsNotNone(vpnservice['id'])
        self.addCleanup(self.admin_client.delete_vpnservice, vpnservice['id'])
        # Assert that created vpnservice is found in API list call
        body = self.client.list_vpnservices()
        vpn_services = [vs['id'] for vs in body['vpnservices']]
        self.assertIn(vpnservice['id'], vpn_services)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('8f33c292-558d-4fdb-b32c-ab1677e8bdc8')
    def test_admin_create_ike_policy_for_tenant(self):
        tenant_id = self._get_tenant_id()

        # Create IKE policy for the newly created tenant
        name = data_utils.rand_name('ike-policy')
        body = (self.admin_client.
                create_ikepolicy(name=name, ike_version="v1",
                                 encryption_algorithm="aes-128",
                                 auth_algorithm="sha256",
                                 tenant_id=tenant_id))
        ikepolicy = body['ikepolicy']
        self.assertIsNotNone(ikepolicy['id'])
        self.addCleanup(self.admin_client.delete_ikepolicy, ikepolicy['id'])

        # Assert that created ike policy is found in API list call
        body = self.client.list_ikepolicies()
        ikepolicies = [ikp['id'] for ikp in body['ikepolicies']]
        self.assertIn(ikepolicy['id'], ikepolicies)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('2997641b-3f2f-4fdf-9af6-bfb9997ecd3b')
    def test_list_vpn_services(self):
        # Verify the VPN service exists in the list of all VPN services
        body = self.client.list_vpnservices()
        vpnservices = body['vpnservices']
        self.assertIn(self.vpnservice['id'], [v['id'] for v in vpnservices])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('8780a28a-deb2-40b0-8433-66f37005c058')
    def test_create_update_delete_vpn_service(self):
        # Creates a VPN service and sets up deletion
        network1 = self.create_network()
        subnet1 = self.create_subnet(network1)
        router1 = self.create_router(data_utils.rand_name('router-'),
                                     external_network_id=self.ext_net_id)
        self.create_router_interface(router1['id'], subnet1['id'])
        name = data_utils.rand_name('vpn-service1')
        body = self.client.create_vpnservice(subnet_id=subnet1['id'],
                                             router_id=router1['id'],
                                             name=name,
                                             admin_state_up=True)
        vpnservice = body['vpnservice']
        self.addCleanup(self._delete_vpn_service, vpnservice['id'])
        # Assert if created vpnservices are not found in vpnservices list
        body = self.client.list_vpnservices()
        vpn_services = [vs['id'] for vs in body['vpnservices']]
        self.assertIsNotNone(vpnservice['id'])
        self.assertIn(vpnservice['id'], vpn_services)

        # TODO(raies): implement logic to update  vpnservice
        # VPNaaS client function to update is implemented.
        # But precondition is that current state of vpnservice
        # should be "ACTIVE" not "PENDING*"

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('631d33ec-8d34-49e1-9e53-576959ea2c57')
    def test_show_vpn_service(self):
        # Verifies the details of a vpn service
        body = self.client.show_vpnservice(self.vpnservice['id'])
        vpnservice = body['vpnservice']
        self.assertEqual(self.vpnservice['id'], vpnservice['id'])
        self.assertEqual(self.vpnservice['name'], vpnservice['name'])
        self.assertEqual(self.vpnservice['description'],
                         vpnservice['description'])
        self.assertEqual(self.vpnservice['router_id'], vpnservice['router_id'])
        self.assertEqual(self.vpnservice['subnet_id'], vpnservice['subnet_id'])
        self.assertEqual(self.vpnservice['tenant_id'], vpnservice['tenant_id'])
        valid_status = ["ACTIVE", "DOWN", "BUILD", "ERROR", "PENDING_CREATE",
                        "PENDING_UPDATE", "PENDING_DELETE"]
        self.assertIn(vpnservice['status'], valid_status)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('b6a665cf-d3df-417f-9477-bd0ef2c56c36')
    def test_list_ike_policies(self):
        # Verify the ike policy exists in the list of all IKE policies
        body = self.client.list_ikepolicies()
        ikepolicies = body['ikepolicies']
        self.assertIn(self.ikepolicy['id'], [i['id'] for i in ikepolicies])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('d9ecb858-7136-4ef5-abba-325c84c57d78')
    def test_create_update_delete_ike_policy(self):
        # Creates a IKE policy
        name = data_utils.rand_name('ike-policy')
        body = (self.client.create_ikepolicy(
                name=name,
                ike_version="v1",
                encryption_algorithm="aes-128",
                auth_algorithm="sha256"))
        ikepolicy = body['ikepolicy']
        self.assertIsNotNone(ikepolicy['id'])
        self.addCleanup(self._delete_ike_policy, ikepolicy['id'])

        # Update IKE Policy
        new_ike = {'name': data_utils.rand_name("New-IKE"),
                   'description': "Updated ike policy",
                   'encryption_algorithm': "aes-256",
                   'ike_version': "v2",
                   'pfs': "group14",
                   'lifetime': {'units': "seconds", 'value': 2000}}
        self.client.update_ikepolicy(ikepolicy['id'], **new_ike)
        # Confirm that update was successful by verifying using 'show'
        body = self.client.show_ikepolicy(ikepolicy['id'])
        ike_policy = body['ikepolicy']
        for key, value in new_ike.items():
            self.assertIn(key, ike_policy)
            self.assertEqual(value, ike_policy[key])

        # Verification of ike policy delete
        self.client.delete_ikepolicy(ikepolicy['id'])
        body = self.client.list_ikepolicies()
        ikepolicies = [ikp['id'] for ikp in body['ikepolicies']]
        self.assertNotIn(ike_policy['id'], ikepolicies)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('5efd625b-41f2-4b80-90e0-81786f3e3309')
    def test_show_ike_policy(self):
        # Verifies the details of a ike policy
        body = self.client.show_ikepolicy(self.ikepolicy['id'])
        ikepolicy = body['ikepolicy']
        self.assertEqual(self.ikepolicy['id'], ikepolicy['id'])
        self.assertEqual(self.ikepolicy['name'], ikepolicy['name'])
        self.assertEqual(self.ikepolicy['description'],
                         ikepolicy['description'])
        self.assertEqual(self.ikepolicy['encryption_algorithm'],
                         ikepolicy['encryption_algorithm'])
        self.assertEqual(self.ikepolicy['auth_algorithm'],
                         ikepolicy['auth_algorithm'])
        self.assertEqual(self.ikepolicy['tenant_id'],
                         ikepolicy['tenant_id'])
        self.assertEqual(self.ikepolicy['pfs'],
                         ikepolicy['pfs'])
        self.assertEqual(self.ikepolicy['phase1_negotiation_mode'],
                         ikepolicy['phase1_negotiation_mode'])
        self.assertEqual(self.ikepolicy['ike_version'],
                         ikepolicy['ike_version'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('c3f7a459-3ccb-42b3-8fab-5ad7a5a791b1')
    def test_list_ipsec_policies(self):
        # Verify the ipsec policy exists in the list of all ipsec policies
        body = self.client.list_ipsecpolicies()
        ipsecpolicies = body['ipsecpolicies']
        self.assertIn(self.ipsecpolicy['id'], [i['id'] for i in ipsecpolicies])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('e3952941-9b9e-48fc-9f36-fc41707a16b4')
    def test_create_update_delete_ipsec_policy(self):
        # Creates an ipsec policy
        ipsec_policy_body = {'name': data_utils.rand_name('ipsec-policy'),
                             'pfs': 'group5',
                             'encryption_algorithm': "aes-128",
                             'auth_algorithm': 'sha256'}
        resp_body = self.client.create_ipsecpolicy(**ipsec_policy_body)
        ipsecpolicy = resp_body['ipsecpolicy']
        self.addCleanup(self._delete_ipsec_policy, ipsecpolicy['id'])
        self._assertExpected(ipsec_policy_body, ipsecpolicy)
        # Verification of ipsec policy update
        new_ipsec = {'description': 'Updated ipsec policy',
                     'pfs': 'group2',
                     'name': data_utils.rand_name("New-IPSec"),
                     'encryption_algorithm': "aes-256",
                     'lifetime': {'units': "seconds", 'value': '2000'}}
        body = self.client.update_ipsecpolicy(ipsecpolicy['id'],
                                              **new_ipsec)
        updated_ipsec_policy = body['ipsecpolicy']
        self._assertExpected(new_ipsec, updated_ipsec_policy)
        # Verification of ipsec policy delete
        self.client.delete_ipsecpolicy(ipsecpolicy['id'])
        self.assertRaises(lib_exc.NotFound,
                          self.client.delete_ipsecpolicy, ipsecpolicy['id'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('86d68b8d-5935-46db-b096-235f70825678')
    def test_show_ipsec_policy(self):
        # Verifies the details of an ipsec policy
        body = self.client.show_ipsecpolicy(self.ipsecpolicy['id'])
        ipsecpolicy = body['ipsecpolicy']
        self._assertExpected(self.ipsecpolicy, ipsecpolicy)

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('95a4a55a-3a10-4f53-9e8a-45be359a245e')
    def test_create_vpnservice_long_name(self):
        """Test excessively long name.

        Without REST checks, this call would return 500 INTERNAL SERVER
        error on internal db failure instead.
        """
        name = _LONG_NAME
        self.assertRaises(
            lib_exc.BadRequest, self.client.create_vpnservice,
            subnet_id=self.subnet['id'], router_id=self.router['id'],
            name=name, admin_state_up=True)

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('6e88b4fb-de5e-464e-97db-a04c68070589')
    def test_create_vpnservice_long_description(self):
        name = data_utils.rand_name('vpn-service1')
        description = _LONG_DESCRIPTION
        self.assertRaises(
            lib_exc.BadRequest, self.client.create_vpnservice,
            subnet_id=self.subnet['id'], router_id=self.router['id'],
            name=name, description=description, admin_state_up=True)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('fdb5a215-0671-449b-91e4-dd58fb223947')
    def test_list_vpn_connections(self):
        # Verify the VPN service exists in the list of all VPN services
        body = self.client.list_ipsec_site_connections()
        ipsec_site_connections = body['ipsec_site_connections']
        self.assertIn(self.ipsec_site_connection['id'],
                      [v['id'] for v in ipsec_site_connections])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('92cfbb15-4299-4fe9-a4e1-f11167e3b057')
    def test_create_delete_vpn_connection_with_legacy_mode(self):
        # Verify create VPN connection
        name = data_utils.rand_name("ipsec_site_connection-")
        body = self.client.create_ipsec_site_connection(
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_cidrs=['10.1.1.0/24', '10.2.2.0/24'],
            name=name,
            mtu=1500,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")
        ipsec_site_connection = body['ipsec_site_connection']
        self.assertEqual(ipsec_site_connection['name'], name)
        self.assertEqual(ipsec_site_connection['mtu'], 1500)
        self.addCleanup(self._delete_ipsec_site_connection,
                        ipsec_site_connection['id'])

        # Verification of IPsec connection delete
        self.client.delete_ipsec_site_connection(ipsec_site_connection['id'])
        body = self.client.list_ipsec_site_connections()
        ipsec_site_connections = body['ipsec_site_connections']
        self.assertNotIn(ipsec_site_connection['id'],
                      [v['id'] for v in ipsec_site_connections])

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('e9977b38-9cd8-4aa6-8bd5-ba15f41c368c')
    def test_create_vpn_connection_missing_peer_cidr(self):
        # Verify create VPN connection with JSON missing peer cidr
        # in legacy mode
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            name=name,
            mtu=1500,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('06a268fa-c53b-4cd1-9350-b83892e4a394')
    def test_create_vpn_service_subnet_not_on_router(self):
        # Verify create VPN service with a subnet not on router
        tenant_id = self._get_tenant_id()

        # Create vpn service for the newly created tenant
        network2 = self.create_network()
        subnet2 = self.create_subnet(network2)
        router2 = self.create_router(data_utils.rand_name('router-'),
                                     external_network_id=self.ext_net_id)
        self.addCleanup(self.admin_client.delete_router, router2['id'])
        self.addCleanup(self.admin_client.delete_network, network2['id'])
        name = data_utils.rand_name('vpn-service')
        self.assertRaises(
            lib_exc.BadRequest,
            self.admin_client.create_vpnservice,
            subnet_id=subnet2['id'],
            router_id=router2['id'],
            name=name,
            admin_state_up=True,
            tenant_id=tenant_id)

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('7678798a-fc20-46cd-ad78-b6b3d599de18')
    def test_create_vpn_connection_small_MTU(self):
        # Verify create VPN connection with small MTU
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_cidrs=['10.1.1.0/24', '10.2.2.0/24'],
            name=name,
            mtu=63,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('d9e1af94-1fbf-4183-b857-82328d4f4b97')
    def test_create_vpn_connection_small_dpd(self):
        # Verify create VPN connection with small dpd
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_cidrs=['10.1.1.0/24', '10.2.2.0/24'],
            name=name,
            dpd=59,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('ea6ae270-9ea4-4446-9022-b7ef7d986762')
    def test_create_vpn_connection_wrong_peer_cidr(self):
        # Verify create VPN connection with wrong peer cidr
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_cidrs=['1.0.0.0/33'],
            name=name,
            mtu=1500,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('68e17a9b-5b33-4d3e-90f2-01a6728d2c26')
    def test_create_connection_with_cidr_and_endpoint_group(self):
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='subnet',
                        endpoints=subnet_id)
        endpoint_group_local = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_local['id'])
        name = data_utils.rand_name('endpoint_group')
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='cidr',
                        endpoints=["10.103.0.0/24", "10.104.0.0/24"])
        endpoint_group_remote = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_remote['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice_no_subnet['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_cidr="10.1.0.0/24",
            peer_ep_group_id=endpoint_group_local['id'],
            local_ep_group_id=endpoint_group_remote['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('d101a6a7-67d3-4418-9d9e-65c4507a9148')
    def test_create_vpn_connection_with_missing_remote_endpoint_group(self):
        # Verify create VPN connection without subnet in vpnservice
        # and has only local endpoint group
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        tenant_id = self._get_tenant_id()
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='subnet',
                        endpoints=subnet_id)
        endpoint_group = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice_no_subnet['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            local_ep_group_id=endpoint_group['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('b5dd8d1e-587b-4e26-84f0-dcb814c65e57')
    def test_create_vpn_connection_with_missing_local_endpoint_group(self):
        # Verify create VPN connection without subnet in vpnservice
        # and only have only local endpoint group
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        tenant_id = self._get_tenant_id()
        name = data_utils.rand_name('endpoint_group')
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='cidr',
                        endpoints=["10.101.0.0/24", "10.102.0.0/24"])
        endpoint_group = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice_no_subnet['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_ep_group_id=endpoint_group['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('427bbc1b-4040-42e4-b661-6395a0bd8762')
    def test_create_connection_with_mix_ip_endpoint_group(self):
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='subnet',
                        endpoints=subnet_id)
        endpoint_group_local = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_local['id'])
        name_v6 = data_utils.rand_name('endpoint_group')
        body_v6 = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name_v6,
                        type='cidr',
                        endpoints=["fec0:101::/64", "fec0:102::/64"])
        endpoint_group_remote = body_v6['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_remote['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertEqual(endpoint_group_local['type'], 'subnet')
        self.assertEqual(endpoint_group_remote['type'], 'cidr')
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice_no_subnet['id'],
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            peer_ep_group_id=endpoint_group_local['id'],
            local_ep_group_id=endpoint_group_remote['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('eac56769-ec2d-4817-ba45-ca101d676bfc')
    def test_create_connection_with_subnet_and_remote_endpoint_group(self):
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        name = data_utils.rand_name('endpoint_group')
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='cidr',
                        endpoints=["10.101.0.0/24", "10.102.0.0/24"])
        endpoint_group = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            peer_ep_group_id=endpoint_group['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('92cd1221-3c7c-47ea-adad-8d8c2fc0f247')
    def test_create_connection_with_subnet_and_local_endpoint_group(self):
        tenant_id = self._get_tenant_id()
        # Create endpoint group for the newly created tenant
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        body = self.client.create_endpoint_group(
                        tenant_id=tenant_id,
                        name=name,
                        type='subnet',
                        endpoints=subnet_id)
        endpoint_group = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Create connections
        name = data_utils.rand_name("ipsec_site_connection-")
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_ipsec_site_connection,
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice['id'],
            peer_address="172.24.4.233",
            local_ep_group_id=endpoint_group['id'],
            name=name,
            admin_state_up=True,
            initiator="bi-directional",
            psk="secret")

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('e0cb784d-ebd9-4ad1-84e1-0cba915784d0')
    def test_create_update_delete_endpoint_group(self):
        # Creates a endpoint-group
        name = data_utils.rand_name('endpoint_group')
        body = (self.client.create_endpoint_group(
                name=name,
                type='cidr',
                endpoints=["10.2.0.0/24", "10.3.0.0/24"]))
        endpoint_group = body['endpoint_group']
        self.assertIsNotNone(endpoint_group['id'])
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Update endpoint-group
        body = {'name': data_utils.rand_name("new_endpoint_group")}
        self.client.update_endpoint_group(endpoint_group['id'],
                        name=name)
        # Confirm that update was successful by verifying using 'show'
        body = self.client.show_endpoint_group(endpoint_group['id'])
        endpoint_group = body['endpoint_group']
        self.assertEqual(name, endpoint_group['name'])
        # Verification of endpoint-group delete
        endpoint_group_id = endpoint_group['id']
        self.client.delete_endpoint_group(endpoint_group['id'])
        body = self.client.list_endpoint_groups()
        endpoint_group = [enp['id'] for enp in body['endpoint_groups']]
        self.assertNotIn(endpoint_group_id, endpoint_group)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('69ee29e9-f72c-4d84-b6ab-c3f5576440fc')
    def test_admin_create_endpoint_group_for_tenant(self):
        # Create endpoint group for the newly created tenant
        tenant_id = self._get_tenant_id()
        name = data_utils.rand_name('endpoint_group')
        body = (self.client.
                create_endpoint_group(
                        name=name,
                        type='cidr',
                        endpoints=["10.2.0.0/24", "10.3.0.0/24"],
                        tenant_id=tenant_id))
        endpoint_group = body['endpoint_group']
        self.assertIsNotNone(endpoint_group['id'])
        self.addCleanup(self._delete_endpoint_group, endpoint_group['id'])
        # Assert that created endpoint group is found in API list call
        endpoint_group_id = endpoint_group['id']
        self.client.delete_endpoint_group(endpoint_group['id'])
        body = self.client.list_endpoint_groups()
        endpoint_group = [enp['id'] for enp in body['endpoint_groups']]
        self.assertNotIn(endpoint_group_id, endpoint_group)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('7a584d9c-0f64-4d39-a66c-261bfa46e369')
    def test_show_endpoint_group(self):
        # Verifies the details of an endpoint group
        body = self.client.show_endpoint_group(self.endpoint_group_local['id'])
        endpoint_group = body['endpoint_group']
        self.assertEqual(self.endpoint_group_local['id'], endpoint_group['id'])
        self.assertEqual(self.endpoint_group_local['name'],
                         endpoint_group['name'])
        self.assertEqual(self.endpoint_group_local['description'],
                         endpoint_group['description'])
        self.assertEqual(self.endpoint_group_local['tenant_id'],
                         endpoint_group['tenant_id'])
        self.assertEqual(self.endpoint_group_local['type'],
                         endpoint_group['type'])
        self.assertEqual(self.endpoint_group_local['endpoints'],
                         endpoint_group['endpoints'])
        # Verifies the details of an endpoint group
        body = self.client.show_endpoint_group(
                         self.endpoint_group_remote['id'])
        endpoint_group = body['endpoint_group']
        # endpoint_group_remote = endpoint_group['id']
        self.assertEqual(self.endpoint_group_remote['id'],
                         endpoint_group['id'])
        self.assertEqual(self.endpoint_group_remote['name'],
                         endpoint_group['name'])
        self.assertEqual(self.endpoint_group_remote['description'],
                         endpoint_group['description'])
        self.assertEqual(self.endpoint_group_remote['tenant_id'],
                         endpoint_group['tenant_id'])
        self.assertEqual(self.endpoint_group_remote['type'],
                         endpoint_group['type'])
        self.assertEqual(self.endpoint_group_remote['endpoints'],
                         endpoint_group['endpoints'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('386f703e-445e-4208-abd2-df66066fd876')
    def test_create_delete_vpn_connection_with_ep_group(self):
        # Creates a endpoint-group with type cidr
        name = data_utils.rand_name('endpoint_group')
        body = self.client.create_endpoint_group(
                name=name,
                type='cidr',
                endpoints=["10.2.0.0/24", "10.3.0.0/24"])
        endpoint_group_remote = body['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_remote['id'])
        # Creates a endpoint-group with type subnet
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        body2 = self.client.create_endpoint_group(
                name=name,
                type='subnet',
                endpoints=subnet_id)
        endpoint_group_local = body2['endpoint_group']
        self.addCleanup(self._delete_endpoint_group,
                        endpoint_group_local['id'])
        # Verify create VPN connection
        name = data_utils.rand_name("ipsec_site_connection-")
        body = self.client.create_ipsec_site_connection(
            ipsecpolicy_id=self.ipsecpolicy['id'],
            ikepolicy_id=self.ikepolicy['id'],
            vpnservice_id=self.vpnservice_no_subnet['id'],
            peer_ep_group_id=endpoint_group_remote['id'],
            local_ep_group_id=endpoint_group_local['id'],
            name=name,
            mtu=1500,
            admin_state_up=True,
            initiator="bi-directional",
            peer_address="172.24.4.233",
            peer_id="172.24.4.233",
            psk="secret")
        ipsec_site_connection = body['ipsec_site_connection']
        self.assertEqual(ipsec_site_connection['name'], name)
        self.assertEqual(ipsec_site_connection['mtu'], 1500)
        self.addCleanup(self._delete_ipsec_site_connection,
                        ipsec_site_connection['id'])

        # Verification of IPsec connection delete
        self.client.delete_ipsec_site_connection(ipsec_site_connection['id'])
        body = self.client.list_ipsec_site_connections()
        ipsec_site_connections = body['ipsec_site_connections']
        self.assertNotIn(ipsec_site_connection['id'],
                      [v['id'] for v in ipsec_site_connections])

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('0ef1b2b0-9b4b-49f1-9473-cb4a0b625543')
    def test_fail_create_endpoint_group_when_wrong_type(self):
        # Creates a endpoint-group with wrong type
        name = data_utils.rand_name('endpoint_group')
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_endpoint_group,
            name=name,
            type='subnet',
            endpoints=["10.2.0.0/24", "10.3.0.0/24"])

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('670844d2-f7b0-466c-8fd7-7d2ad6e832ee')
    def test_fail_create_endpoint_group_when_provide_subnet_id_with_cidr(self):
        # Creates a endpoint-group when provide subnet id with type cidr
        name = data_utils.rand_name('endpoint_group')
        subnet_id = self.subnet['id']
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_endpoint_group,
            name=name,
            type='cidr',
            endpoints=subnet_id)

    @decorators.attr(type=['negative', 'smoke'])
    @decorators.idempotent_id('d7516513-f2a2-42bd-8cea-baba73b93a22')
    def test_fail_create_endpoint_group_with_mixed_IP_version(self):
        # Creates a endpoint-group with mixed IP version
        name = data_utils.rand_name('endpoint_group')
        self.assertRaises(
            lib_exc.BadRequest,
            self.client.create_endpoint_group,
            name=name,
            type='cidr',
            endpoints=["10.2.0.0/24", "2000::1"])



# Copyright (c) 2015 Ericsson.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions
from testtools import ExpectedException

from neutron_tempest_plugin.bgpvpn.base import BaseBgpvpnTest as base


class BgpvpnTest(base):
    """Tests the following operations in the Neutron API:

        create bgpvpn
        delete bgpvpn
        show bgpvpn
        list bgpvpns
        associate network to bgpvpn
        disassociate network from bgpvpn
        show network association
        list network associations
        update route targets

    v2.0 of the Neutron API is assumed. It is also assumed that the following
    options are defined in the [network] section of etc/tempest.conf:

    ...
    """

    @decorators.idempotent_id('4f90deb2-eb8e-4e7d-9d68-c5b5cc657f7e')
    def test_create_bgpvpn(self):
        self.create_bgpvpn(self.bgpvpn_admin_client)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('0a911d61-d908-4c21-a11e-e403ac0d8e38')
    def test_create_bgpvpn_as_non_admin_fail(self):
        self.assertRaises(exceptions.Forbidden,
                          self.create_bgpvpn, self.bgpvpn_client)

    @decorators.idempotent_id('709b23b0-9719-47df-9f53-b0812a5d5a48')
    def test_delete_bgpvpn(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        self.delete_bgpvpn(self.bgpvpn_admin_client, bgpvpn)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('596abfc2-fd89-491d-863d-25459db1df4b')
    def test_delete_bgpvpn_as_non_admin_fail(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        self.assertRaises(exceptions.Forbidden,
                          self.bgpvpn_client.delete_bgpvpn, bgpvpn['id'])

    @decorators.idempotent_id('9fa29db8-35d0-4beb-a986-23c369499ab1')
    def test_show_bgpvpn(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        bgpvpn_details = self.bgpvpn_client.show_bgpvpn(bgpvpn['id'])['bgpvpn']
        self.assertEqual(bgpvpn['id'], bgpvpn_details['id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('b20110bb-393b-4342-8b30-6486cd2b4fc6')
    def test_show_bgpvpn_as_non_owner_fail(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_alt_client.show_bgpvpn, bgpvpn['id'])

    @decorators.idempotent_id('7a7feca2-1c24-4f5d-ad4b-b0e5a712adb1')
    def test_list_bgpvpn(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        bgpvpns = self.bgpvpn_client.list_bgpvpns()['bgpvpns']
        self.assertIn(bgpvpn['id'],
                      [bgpvpn_alt['id'] for bgpvpn_alt in bgpvpns])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('4875e65d-0b65-40c0-9efd-309420686ab4')
    def test_list_bgpvpn_as_non_owner_fail(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        bgpvpns_alt = self.bgpvpn_alt_client.list_bgpvpns()['bgpvpns']
        self.assertNotIn(bgpvpn['id'],
                         [bgpvpn_alt['id'] for bgpvpn_alt in bgpvpns_alt])

    @decorators.idempotent_id('096281da-356d-4c04-bd55-784a26bb1b0c')
    def test_list_show_network_association(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        network = self.networks_client.create_network()['network']

        association = self.bgpvpn_client.create_network_association(
            bgpvpn['id'], network['id'])['network_association']
        net_assocs = self.bgpvpn_client\
            .list_network_associations(bgpvpn['id'])['network_associations']
        self.assertIn(association['id'],
                      [net_assoc['id'] for net_assoc in net_assocs])
        net_assoc_details = self.bgpvpn_client\
            .show_network_association(bgpvpn['id'],
                                      association['id'])['network_association']
        self.assertEqual(association['id'], net_assoc_details['id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('57b0da93-8e37-459f-9aaf-f903acc36025')
    def test_show_netassoc_as_non_owner_fail(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        network = self.networks_client.create_network()['network']

        net_assoc = self.bgpvpn_client.create_network_association(
            bgpvpn['id'], network['id'])['network_association']

        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_alt_client.show_network_association,
                          bgpvpn['id'], net_assoc['id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('2cbb10af-bf9c-4b32-b6a6-4066de783758')
    def test_list_netassoc_as_non_owner_fail(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        network = self.networks_client.create_network()['network']

        self.bgpvpn_client.create_network_association(bgpvpn['id'],
                                                      network['id'])
        net_assocs_alt = self.bgpvpn_alt_client\
            .list_network_associations(bgpvpn['id'])
        self.assertFalse(net_assocs_alt['network_associations'])

    @decorators.idempotent_id('51e1b079-aefa-4c37-8b1a-0567b3ef7954')
    def test_associate_disassociate_network(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        network = self.networks_client.create_network()
        network_id = network['network']['id']

        # Associate the network to the bgpvpn resource
        association = self.bgpvpn_client.create_network_association(
            bgpvpn['id'], network_id)
        self.assertEqual(association['network_association']['network_id'],
                         network_id)
        updated_bgpvpn = self.bgpvpn_client.show_bgpvpn(bgpvpn['id'])
        self.assertEqual(updated_bgpvpn['bgpvpn']['networks'], [network_id])

        # Disassociate the network from the bgpvpn resource
        self.bgpvpn_client.delete_network_association(
            bgpvpn['id'],
            association['network_association']['id'])
        updated_bgpvpn = self.bgpvpn_client.show_bgpvpn(bgpvpn['id'])
        self.assertEqual(updated_bgpvpn['bgpvpn']['networks'], [])

        self.networks_client.delete_network(network_id)

    @decorators.idempotent_id('559013fd-1e34-4fde-9599-f8aafe9ae716')
    def test_update_route_target(self):
        bgpvpn = self.create_bgpvpn(
            self.bgpvpn_admin_client,
            route_targets=['64512:1'],
            import_targets=['64512:2'],
            export_targets=['64512:3'])
        bgpvpn = self.bgpvpn_admin_client.update_bgpvpn(
            bgpvpn['id'],
            route_targets=['64512:4'],
            import_targets=['64512:5'],
            export_targets=['64512:6']
        )['bgpvpn']
        self.assertEqual(['64512:4'], bgpvpn['route_targets'])
        self.assertEqual(['64512:5'], bgpvpn['import_targets'])
        self.assertEqual(['64512:6'], bgpvpn['export_targets'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('e35eb9be-fe1f-406c-b36b-fc1879328313')
    def test_update_route_target_non_admin_fail(self):
        bgpvpn = self.create_bgpvpn(
            self.bgpvpn_admin_client,
            tenant_id=self.bgpvpn_client.project_id,
            route_targets=['64512:1'])
        with ExpectedException(exceptions.Forbidden):
            self.bgpvpn_client.update_bgpvpn(
                bgpvpn['id'],
                route_targets=['64512:2'],
                import_targets=['64512:3'],
                export_targets=['64512:4'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('464ca6f9-86e4-4ee3-9c65-f1edae93223d')
    def test_create_bgpvpn_with_invalid_routetargets(self):
        """Create a bgpvpn with invalid route target

        This test verifies that invalid route targets,import targets,
        export targets are rejected by the Create API
        """
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
            "route_targets": ["0"]
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.create_bgpvpn, **postdata)
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
            "import_targets": ["test", " "]
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.create_bgpvpn, **postdata)
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
            "export_targets": ["64512:1000000000000", "xyz"]
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.create_bgpvpn, **postdata)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('7d4e9b87-e1ab-47a7-a8d6-9d179365556a')
    def test_update_bgpvpn_invalid_routetargets(self):
        """Update the bgpvpn with invalid route targets

        This test  verifies that invalid  route targets,import targets
        and export targets are rejected by the Update API
        """
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
        }
        bgpvpn = self.bgpvpn_admin_client.create_bgpvpn(**postdata)
        updatedata = {
            "route_targets": ["0"]
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.update_bgpvpn,
                          bgpvpn['bgpvpn']['id'], **updatedata)
        updatedata = {
            "import_targets": ["test", " "]
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.update_bgpvpn,
                          bgpvpn['bgpvpn']['id'], **updatedata)
        updatedata = {
            "export_targets": ["64512:1000000000000", "xyz"],
        }
        self.assertRaises(exceptions.BadRequest,
                          self.bgpvpn_admin_client.update_bgpvpn,
                          bgpvpn['bgpvpn']['id'], **updatedata)

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('f049ce21-d239-47c0-b13f-fb57a2a558ce')
    def test_associate_invalid_network(self):
        """Associate the invalid network in bgpvpn

        This test verifies that invalid network id,bgpvpn id
        are rejected by the associate API
        """
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
        }
        bgpvpn = self.bgpvpn_admin_client.create_bgpvpn(**postdata)
        network = self.networks_client.create_network()
        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_client.create_network_association,
                          bgpvpn['bgpvpn']['id'], uuidutils.generate_uuid())
        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_client.create_network_association,
                          uuidutils.generate_uuid(),
                          network['network']['id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('078b2660-4adb-4c4c-abf0-b77bf0bface5')
    def test_disassociate_invalid_network(self):
        """Disassociate the invalid network in bgpvpn

        This test verifies that invalid network id,
        bgpvpn id are rejected by the disassociate API
        """
        postdata = {
            "name": "testbgpvpn",
            "tenant_id": self.bgpvpn_client.project_id,
        }
        bgpvpn = self.bgpvpn_admin_client.create_bgpvpn(**postdata)
        network = self.networks_client.create_network()
        association = self.bgpvpn_client.create_network_association(
            bgpvpn['bgpvpn']['id'], network['network']['id'])
        self.assertEqual(association['network_association'][
                         'network_id'], network['network']['id'])
        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_client.delete_network_association,
                          bgpvpn['bgpvpn']['id'],
                          uuidutils.generate_uuid())
        self.assertRaises(exceptions.NotFound,
                          self.bgpvpn_client.delete_network_association,
                          uuidutils.generate_uuid(),
                          association['network_association']['id'])

    @decorators.idempotent_id('de8d94b0-0239-4a48-9574-c3a4a4f7cacb')
    def test_associate_disassociate_router(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        router = self.routers_client.create_router()
        router_id = router['router']['id']

        # Associate the network to the bgpvpn resource
        association = self.bgpvpn_client.create_router_association(
            bgpvpn['id'], router_id)
        self.assertEqual(association['router_association']['router_id'],
                         router_id)
        updated_bgpvpn = self.bgpvpn_client.show_bgpvpn(bgpvpn['id'])
        self.assertEqual(updated_bgpvpn['bgpvpn']['routers'], [router_id])

        # Disassociate the network from the bgpvpn resource
        self.bgpvpn_client.delete_router_association(
            bgpvpn['id'],
            association['router_association']['id'])
        updated_bgpvpn = self.bgpvpn_client.show_bgpvpn(bgpvpn['id'])
        self.assertEqual(updated_bgpvpn['bgpvpn']['routers'], [])

        self.routers_client.delete_router(router_id)

    @decorators.idempotent_id('3ae91755-b1b6-4c62-a699-a44eeb4ee522')
    def test_list_show_router_association(self):
        bgpvpn = self.create_bgpvpn(self.bgpvpn_admin_client,
                                    tenant_id=self.bgpvpn_client.project_id)
        router = self.routers_client.create_router()
        router_id = router['router']['id']

        association = self.bgpvpn_client.create_router_association(
            bgpvpn['id'], router_id)['router_association']
        rtr_assocs = self.bgpvpn_client\
            .list_router_associations(bgpvpn['id'])['router_associations']
        self.assertIn(association['id'],
                      [rtr_assoc['id'] for rtr_assoc in rtr_assocs])
        rtr_assoc_details = self.bgpvpn_client\
            .show_router_association(bgpvpn['id'],
                                     association['id'])['router_association']
        self.assertEqual(association['id'], rtr_assoc_details['id'])

    @decorators.attr(type=['negative'])
    @decorators.idempotent_id('4be1f073-fe57-4858-b7b9-9a189e90b770')
    def test_attach_associated_subnet_to_associated_router(self):
        # Create a first bgpvpn and associate a network with a subnet to it
        bgpvpn_net = self.create_bgpvpn(
            self.bgpvpn_admin_client,
            tenant_id=self.bgpvpn_client.project_id)
        network = self.create_network()
        subnet = self.create_subnet(network)
        self.bgpvpn_client.create_network_association(
            bgpvpn_net['id'], network['id'])

        # Create a second bgpvpn and associate a router to it
        bgpvpn_router = self.create_bgpvpn(
            self.bgpvpn_admin_client,
            tenant_id=self.bgpvpn_client.project_id)

        router = self.create_router(
            router_name=data_utils.rand_name('test-bgpvpn-'))
        self.bgpvpn_client.create_router_association(
            bgpvpn_router['id'],
            router['id'])

        # Attach the subnet of the network to the router
        subnet_data = {'subnet_id': subnet['id']}
        self.assertRaises(exceptions.Conflict,
                          self.routers_client.add_router_interface,
                          router['id'],
                          **subnet_data)



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import constants
import testtools

from oslo_log import log
from tempest.common import utils as tempest_utils
from tempest.common import waiters
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators

from neutron_tempest_plugin.common import ip
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin import exceptions
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants as const

CONF = config.CONF
LOG = log.getLogger(__name__)
EPHEMERAL_PORT_RANGE = {'min': 32768, 'max': 65535}


def get_capture_script(interface, tcp_port, packet_types, result_file):
    return """#!/bin/bash
tcpdump -i %(interface)s -vvneA -s0 -l -c1 \
"dst port %(port)s and tcp[tcpflags] == %(packet_types)s" &> %(result)s &
    """ % {'interface': interface,
           'port': tcp_port,
           'packet_types': packet_types,
           'result': result_file}


class BaseNetworkSecGroupTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']
    required_extensions = ['router', 'security-group']

    def _test_connection_and_log(self, con, *args, **kwargs):
        try:
            con.test_connection(*args, **kwargs)
        except utils.WaitTimeout:
            self._log_console_output()
            self._log_local_network_status()
            raise

    @classmethod
    def setup_credentials(cls):
        super().setup_credentials()
        cls.network_client = cls.os_admin.network_client

    @classmethod
    def setup_clients(cls):
        super().setup_clients()
        cls.project_id = cls.os_primary.credentials.tenant_id

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        # setup basic topology for servers we can log into it
        cls.reserve_external_subnet_cidrs()
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        if cls.ipv6_mode:
            cls.subnet_v6 = cls.create_subnet(
                cls.network,
                ip_version=constants.IP_VERSION_6,
                ipv6_ra_mode=cls.ipv6_mode,
                ipv6_address_mode=cls.ipv6_mode)
            cls.create_router_interface(cls.router['id'], cls.subnet_v6['id'])
        cls.keypair = cls.create_keypair()

    def setUp(self):
        super().setUp()
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.network_client.reset_quotas, self.project_id)
        self.network_client.update_quotas(self.project_id, security_group=-1)
        self.network_client.update_quotas(self.project_id,
                                          security_group_rule=-1)

    def create_vm_testing_sec_grp(self, num_servers=2, security_groups=None,
                                  ports=None, network_id=None,
                                  use_advanced_image=False):
        """Create instance for security group testing

        :param num_servers (int): number of servers to spawn
        :param security_groups (list): list of security groups
        :param ports* (list): list of ports
        :param: use_advanced_image (bool): use Cirros (False) or
                advanced guest image
        *Needs to be the same length as num_servers
        """
        if (not use_advanced_image or
                CONF.neutron_plugin_options.default_image_is_advanced):
            flavor_ref = CONF.compute.flavor_ref
            image_ref = CONF.compute.image_ref
            username = CONF.validation.image_ssh_user
        else:
            flavor_ref = CONF.neutron_plugin_options.advanced_image_flavor_ref
            image_ref = CONF.neutron_plugin_options.advanced_image_ref
            username = CONF.neutron_plugin_options.advanced_image_ssh_user
        network_id = network_id or self.network['id']
        servers, fips, server_ssh_clients = ([], [], [])
        for i in range(num_servers):
            server_args = {
                'flavor_ref': flavor_ref,
                'image_ref': image_ref,
                'key_name': self.keypair['name'],
                'networks': [{'uuid': network_id}],
                'security_groups': security_groups
            }
            if ports is not None:
                server_args['networks'][0].update({'port': ports[i]['id']})
            servers.append(self.create_server(**server_args))
        for i, server in enumerate(servers):
            waiters.wait_for_server_status(
                self.os_primary.servers_client, server['server']['id'],
                const.SERVER_STATUS_ACTIVE)
            port = self.client.list_ports(
                network_id=network_id, device_id=server['server'][
                    'id'])['ports'][0]
            fips.append(self.create_floatingip(port=port))
            server_ssh_clients.append(ssh.Client(
                fips[i]['floating_ip_address'], username,
                pkey=self.keypair['private_key']))
        return server_ssh_clients, fips, servers

    def _get_default_security_group(self):
        sgs = self.os_primary.network_client.list_security_groups(
            project_id=self.project_id)['security_groups']
        for sg in sgs:
            if sg['name'] == 'default':
                return sg

    def _create_security_group(self, name_prefix, **kwargs):
        if self.stateless_sg:
            kwargs['stateful'] = False
        return super().create_security_group(
            name=data_utils.rand_name(name_prefix), **kwargs)

    def _create_client_and_server_vms(
            self, allowed_tcp_port=None, use_advanced_image=False):
        networks = {
            'server': self.network,
            'client': self.create_network()}
        subnet = self.create_subnet(networks['client'])
        self.create_router_interface(self.router['id'], subnet['id'])

        security_groups = {}
        for sg_name in ["server", "client"]:
            sg = self._create_security_group('vm_%s_secgrp' % sg_name)
            self.create_loginable_secgroup_rule(
                secgroup_id=sg['id'])
            if allowed_tcp_port:
                self.create_security_group_rule(
                    security_group_id=sg['id'],
                    protocol=constants.PROTO_NAME_TCP,
                    direction=constants.INGRESS_DIRECTION,
                    port_range_min=allowed_tcp_port,
                    port_range_max=allowed_tcp_port)
            else:
                self.create_pingable_secgroup_rule(sg['id'])
            if self.stateless_sg:
                self.create_ingress_metadata_secgroup_rule(
                    secgroup_id=sg['id'])
            security_groups[sg_name] = sg
        # NOTE(slaweq): we need to iterate over create_vm_testing_sec_grp as
        # this method plugs all SGs to all VMs and we need each vm to use other
        # SGs
        ssh_clients = {}
        fips = {}
        servers = {}
        for server_name, sg in security_groups.items():
            _ssh_clients, _fips, _servers = self.create_vm_testing_sec_grp(
                num_servers=1,
                security_groups=[{'name': sg['name']}],
                network_id=networks[server_name]['id'],
                use_advanced_image=use_advanced_image)
            ssh_clients[server_name] = _ssh_clients[0]
            fips[server_name] = _fips[0]
            servers[server_name] = _servers[0]
        return ssh_clients, fips, servers, security_groups

    def _test_connectivity_between_vms_using_different_sec_groups(self):
        TEST_TCP_PORT = 1022
        ssh_clients, fips, servers, security_groups = (
            self._create_client_and_server_vms(TEST_TCP_PORT))

        # make sure tcp connectivity between vms works fine
        for fip in fips.values():
            self.check_connectivity(fip['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])
        # Check connectivity between servers
        def _message_received(server_ssh_client, client_ssh_client,
                              dest_fip, servers):
            expected_msg = "Test_msg"
            utils.kill_nc_process(server_ssh_client)
            self.nc_listen(server_ssh_client,
                           TEST_TCP_PORT,
                           constants.PROTO_NAME_TCP,
                           expected_msg,
                           list(servers.values()))
            try:
                received_msg = self.nc_client(
                    dest_fip,
                    TEST_TCP_PORT,
                    constants.PROTO_NAME_TCP,
                    ssh_client=client_ssh_client)
                return received_msg and expected_msg in received_msg
            except exceptions.ShellCommandFailed:
                return False

        if self.stateless_sg:
            # In case of stateless SG connectivity will not work without
            # explicit allow ingress response from server to client
            utils.wait_until_true(
                lambda: not _message_received(
                    ssh_clients['server'], ssh_clients['client'],
                    fips['server']['fixed_ip_address'], servers))
            self.create_security_group_rule(
                security_group_id=security_groups['client']['id'],
                protocol=constants.PROTO_NAME_TCP,
                direction=constants.INGRESS_DIRECTION,
                port_range_min=EPHEMERAL_PORT_RANGE['min'],
                port_range_max=EPHEMERAL_PORT_RANGE['max'])

        utils.wait_until_true(
            lambda: _message_received(
                ssh_clients['server'], ssh_clients['client'],
                fips['server']['fixed_ip_address'], servers))

    def _test_ip_prefix(self, rule_list, should_succeed):
        # Add specific remote prefix to VMs and check connectivity
        ssh_secgrp = self._create_security_group('ssh_secgrp')
        self.create_loginable_secgroup_rule(
            secgroup_id=ssh_secgrp['id'])
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=ssh_secgrp['id'])
        icmp_secgrp = self._create_security_group('icmp_secgrp')
        self.create_secgroup_rules(
            rule_list, secgroup_id=icmp_secgrp['id'])
        for sec_grp in (ssh_secgrp, icmp_secgrp):
            self.security_groups.append(sec_grp)
        security_groups_list = [
            {'name': ssh_secgrp['name']},
            {'name': icmp_secgrp['name']}]
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            security_groups=security_groups_list)

        # make sure ssh connectivity works
        self.check_connectivity(fips[0]['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

        # make sure ICMP connectivity works
        self.check_remote_connectivity(server_ssh_clients[0], fips[1][
            'fixed_ip_address'], should_succeed=should_succeed,
            servers=servers)

    def _test_default_sec_grp_scenarios(self):
        # Ensure that SG used in tests is stateful or stateless as required
        default_sg_id = self._get_default_security_group()['id']
        self.os_primary.network_client.update_security_group(
            default_sg_id, stateful=not self.stateless_sg)
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=default_sg_id)
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp()

        # Check ssh connectivity when you add sec group rule, enabling ssh
        self.create_loginable_secgroup_rule(default_sg_id)
        self.check_connectivity(fips[0]['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

        # make sure ICMP connectivity still fails as only ssh rule was added
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # Check ICMP connectivity between VMs without specific rule for that
        # It should work though the rule is not configured
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            servers=servers)

        # Check ICMP connectivity from VM to external network
        if self.stateless_sg:
            # NOTE(slaweq): in case of stateless SG explicit ingress rule for
            # the ICMP replies needs to be added too
            self.create_pingable_secgroup_rule(default_sg_id)
        subnets = self.os_admin.network_client.list_subnets(
            network_id=CONF.network.public_network_id)['subnets']
        ext_net_ip = None
        for subnet in subnets:
            if subnet['ip_version'] == 4:
                ext_net_ip = subnet['gateway_ip']
                break
        self.assertTrue(ext_net_ip)
        self.check_remote_connectivity(server_ssh_clients[0], ext_net_ip,
                                       servers=servers)
        return server_ssh_clients, fips, servers

    def _test_protocol_number_rule(self):
        # protocol number is added instead of str in security rule creation
        name = data_utils.rand_name("test_protocol_number_rule")
        security_group = self._create_security_group(name)
        port = self.create_port(network=self.network, name=name,
                                security_groups=[security_group['id']])
        _, fips, _ = self.create_vm_testing_sec_grp(num_servers=1,
                                                    ports=[port])
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_ip_prefix': '0.0.0.0/0'}]
        self.create_secgroup_rules(rule_list, secgroup_id=security_group['id'])
        self.ping_ip_address(fips[0]['floating_ip_address'])

    def _test_two_sec_groups(self):
        # add 2 sec groups to VM and test rules of both are working
        ssh_secgrp = self._create_security_group('ssh_secgrp')
        self.create_loginable_secgroup_rule(
            secgroup_id=ssh_secgrp['id'])
        icmp_secgrp = self._create_security_group('icmp_secgrp')
        self.create_pingable_secgroup_rule(
            secgroup_id=icmp_secgrp['id'])
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=ssh_secgrp['id'])
        for sec_grp in (ssh_secgrp, icmp_secgrp):
            self.security_groups.append(sec_grp)
        security_groups_list = [
            {'name': ssh_secgrp['name']},
            {'name': icmp_secgrp['name']}]
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            num_servers=1, security_groups=security_groups_list)
        # make sure ssh connectivity works
        self.check_connectivity(fips[0]['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])
        # make sure ICMP connectivity works
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=True)
        ports = self.client.list_ports(device_id=servers[0]['server']['id'])
        port_id = ports['ports'][0]['id']

        # update port with ssh security group only
        self.os_primary.network_client.update_port(
            port_id, security_groups=[ssh_secgrp['id']])

        # make sure ssh connectivity works
        self.check_connectivity(fips[0]['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

        # make sure ICMP connectivity doesn't work
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # update port with ssh and ICMP security groups
        self.os_primary.network_client.update_port(
            port_id, security_groups=[icmp_secgrp['id'], ssh_secgrp['id']])

        # make sure ssh connectivity  works after update
        self.check_connectivity(fips[0]['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

        # make sure ICMP connectivity works after update
        self.ping_ip_address(fips[0]['floating_ip_address'])

    def _test_remote_group(self):
        # create a new sec group
        ssh_secgrp = self._create_security_group('ssh_secgrp')
        # configure sec group to support SSH connectivity
        self.create_loginable_secgroup_rule(
            secgroup_id=ssh_secgrp['id'])
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=ssh_secgrp['id'])
        # spawn two instances with the sec group created
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            security_groups=[{'name': ssh_secgrp['name']}])
        # verify SSH functionality
        for i in range(2):
            self.check_connectivity(fips[i]['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])
        # try to ping instances without ICMP permissions
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            should_succeed=False)
        # add ICMP support to the remote group
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_group_id': ssh_secgrp['id']}]
        self.create_secgroup_rules(
            rule_list, secgroup_id=ssh_secgrp['id'])
        # verify ICMP connectivity between instances works
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            servers=servers)
        # make sure ICMP connectivity doesn't work from framework
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

    def _test_remote_group_and_remote_address_group(self):
        """Test SG rules with remote group and remote address group

        This test checks the ICMP connection among two servers using a security
        group rule with remote group and another rule with remote address
        group. The connection should be granted when at least one of the rules
        is applied. When both rules are applied (overlapped), removing one of
        them should not disable the connection.
        """
        # create a new sec group
        ssh_secgrp = self._create_security_group('ssh_secgrp')
        # configure sec group to support SSH connectivity
        self.create_loginable_secgroup_rule(
            secgroup_id=ssh_secgrp['id'])
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=ssh_secgrp['id'])
        # spawn two instances with the sec group created
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            security_groups=[{'name': ssh_secgrp['name']}])
        # verify SSH functionality
        for i in range(2):
            self.check_connectivity(fips[i]['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])
        # try to ping instances without ICMP permissions
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            should_succeed=False)
        # add ICMP support to the remote group
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_group_id': ssh_secgrp['id']}]
        remote_sg_rid = self.create_secgroup_rules(
            rule_list, secgroup_id=ssh_secgrp['id'])[0]['id']
        # verify ICMP connectivity between instances works
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            servers=servers)
        # make sure ICMP connectivity doesn't work from framework
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # add ICMP rule with remote address group
        address_set = [str(netaddr.IPNetwork(fips[0]['fixed_ip_address']))]
        if self.stateless_sg:
            address_set.append(
                str(netaddr.IPNetwork(fips[1]['fixed_ip_address'])))
        test_ag = self.create_address_group(
            name=data_utils.rand_name('test_ag'),
            addresses=address_set)
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_address_group_id': test_ag['id']}]
        remote_ag_rid = self.create_secgroup_rules(
            rule_list, secgroup_id=ssh_secgrp['id'])[0]['id']
        # verify ICMP connectivity between instances still works
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            servers=servers)
        # make sure ICMP connectivity doesn't work from framework
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # Remove the ICMP rule with remote group
        self.client.delete_security_group_rule(remote_sg_rid)
        # verify ICMP connectivity between instances still works as granted
        # by the rule with remote address group
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            servers=servers)
        # make sure ICMP connectivity doesn't work from framework
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # Remove the ICMP rule with remote address group
        self.client.delete_security_group_rule(remote_ag_rid)
        # verify ICMP connectivity between instances doesn't work now
        self.check_remote_connectivity(
            server_ssh_clients[0], fips[1]['fixed_ip_address'],
            should_succeed=False)
        # make sure ICMP connectivity doesn't work from framework
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

    def _test_multiple_ports_secgroup_inheritance(self):
        """Test multiple port security group inheritance

        This test creates two ports with security groups, then
        boots two instances and verify that the security group was
        inherited properly and enforced in these instances.
        """
        # create a security group and make it loginable and pingable
        secgrp = self._create_security_group('secgrp')
        self.create_loginable_secgroup_rule(
            secgroup_id=secgrp['id'])
        self.create_pingable_secgroup_rule(
            secgroup_id=secgrp['id'])
        if self.stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=secgrp['id'])
        # create two ports with fixed IPs and the security group created
        ports = []
        for i in range(2):
            ports.append(self.create_port(
                self.network, fixed_ips=[{'subnet_id': self.subnets[0]['id']}],
                security_groups=[secgrp['id']]))
        # spawn instances with the ports created
        server_ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            ports=ports)
        # verify ICMP reachability and ssh connectivity
        for fip in fips:
            self.ping_ip_address(fip['floating_ip_address'])
            self.check_connectivity(fip['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])

    def _test_multiple_ports_portrange_remote(self):
        initial_security_groups = []
        if self.stateless_sg:
            md_secgrp = self._create_security_group('metadata_secgrp')
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=md_secgrp['id'])
            initial_security_groups.append(
                {'name': md_secgrp['name']})

        ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            num_servers=3, security_groups=initial_security_groups)
        secgroups = []
        ports = []

        # Create remote and test security groups
        for i in range(0, 2):
            secgroups.append(
                self._create_security_group('secgrp-%d' % i))
            # configure sec groups to support SSH connectivity
            self.create_loginable_secgroup_rule(
                secgroup_id=secgroups[-1]['id'])

        # Configure security groups, first two servers as remotes
        for i, server in enumerate(servers):
            port = self.client.list_ports(
                network_id=self.network['id'], device_id=server['server'][
                    'id'])['ports'][0]
            ports.append(port)
            secgroup = secgroups[0 if i in range(0, 2) else 1]
            self.client.update_port(port['id'], security_groups=[
                secgroup['id']])

        # verify SSH functionality
        for fip in fips:
            self.check_connectivity(fip['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])

        test_ip = ports[2]['fixed_ips'][0]['ip_address']

        # verify that conections are not working
        for port in range(80, 84):
            with utils.StatefulConnection(
                    ssh_clients[0], ssh_clients[2], test_ip, port) as con:
                self._test_connection_and_log(con, should_pass=False)

        # add two remote-group rules with port-ranges
        rule_list = [{'protocol': constants.PROTO_NUM_TCP,
                      'direction': constants.INGRESS_DIRECTION,
                      'port_range_min': '80',
                      'port_range_max': '81',
                      'remote_group_id': secgroups[0]['id']},
                     {'protocol': constants.PROTO_NUM_TCP,
                      'direction': constants.INGRESS_DIRECTION,
                      'port_range_min': '82',
                      'port_range_max': '83',
                      'remote_group_id': secgroups[0]['id']}]
        if self.stateless_sg:
            rule_list.append({
                'protocol': constants.PROTO_NUM_TCP,
                'direction': constants.EGRESS_DIRECTION,
                'remote_group_id': secgroups[0]['id']})
            # NOTE(slaweq): in case of stateless SG, client needs to have also
            # rule which will explicitly accept ingress connections from
            # secgroup[1]

            self.create_security_group_rule(
                security_group_id=secgroups[0]['id'],
                protocol=constants.PROTO_NAME_TCP,
                direction=constants.INGRESS_DIRECTION,
                remote_group_id=secgroups[1]['id'])

        self.create_secgroup_rules(
            rule_list, secgroup_id=secgroups[1]['id'])

        # verify that conections are working
        for port in range(80, 84):
            with utils.StatefulConnection(
                    ssh_clients[0], ssh_clients[2], test_ip, port) as con:
                self._test_connection_and_log(con)

        # list the tcp rule id by SG id and port-range
        sg_rule_id = self.os_primary.network_client.list_security_group_rules(
            security_group_id=secgroups[1]['id'],
            port_range_min=80)['security_group_rules'][0]['id']

        # delete the tcp rule from the security group
        self.client.delete_security_group_rule(sg_rule_id)

        # verify that conections are not working
        for port in range(80, 82):
            with utils.StatefulConnection(
                    ssh_clients[0], ssh_clients[2], test_ip, port) as con:
                self._test_connection_and_log(con, should_pass=False)

    def _test_overlapping_sec_grp_rules(self):
        """Test security group rules with overlapping port ranges"""
        initial_security_groups = []
        if self.stateless_sg:
            md_secgrp = self._create_security_group('metadata_secgrp')
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=md_secgrp['id'])
            initial_security_groups.append(
                {'name': md_secgrp['name']})
        client_ssh, _, vms = self.create_vm_testing_sec_grp(
            num_servers=2, security_groups=initial_security_groups)
        tmp_ssh, _, tmp_vm = self.create_vm_testing_sec_grp(
            num_servers=1, security_groups=initial_security_groups)
        srv_ssh = tmp_ssh[0]
        srv_vm = tmp_vm[0]
        srv_port = self.client.list_ports(network_id=self.network['id'],
                device_id=srv_vm['server']['id'])['ports'][0]
        srv_ip = srv_port['fixed_ips'][0]['ip_address']
        secgrps = []
        for i, vm in enumerate(vms):
            sg = self._create_security_group('secgrp-%d' % i)
            self.create_loginable_secgroup_rule(secgroup_id=sg['id'])
            port = self.client.list_ports(network_id=self.network['id'],
                    device_id=vm['server']['id'])['ports'][0]
            self.client.update_port(port['id'], security_groups=[sg['id']])
            secgrps.append(sg)
        tcp_port = 3000
        rule_list = [{'protocol': constants.PROTO_NUM_TCP,
                      'direction': constants.INGRESS_DIRECTION,
                      'port_range_min': tcp_port,
                      'port_range_max': tcp_port,
                      'remote_group_id': secgrps[0]['id']},
                     {'protocol': constants.PROTO_NUM_TCP,
                      'direction': constants.INGRESS_DIRECTION,
                      'port_range_min': tcp_port,
                      'port_range_max': tcp_port + 2,
                      'remote_group_id': secgrps[1]['id']}]
        self.client.update_port(srv_port['id'],
                security_groups=[secgrps[0]['id'], secgrps[1]['id']])
        self.create_secgroup_rules(rule_list, secgroup_id=secgrps[0]['id'])

        if self.stateless_sg:
            # NOTE(slaweq): in case of stateless SG, client needs to have also
            # rule which will explicitly accept ingress TCP connections which
            # will be replies from the TCP server so it will use random
            # destination port (depends on the src port choosen by client while
            # establishing connection)
            self.create_security_group_rule(
                security_group_id=secgrps[0]['id'],
                protocol=constants.PROTO_NAME_TCP,
                direction=constants.INGRESS_DIRECTION)
            self.create_security_group_rule(
                security_group_id=secgrps[1]['id'],
                protocol=constants.PROTO_NAME_TCP,
                direction=constants.INGRESS_DIRECTION)

        # The conntrack entries are ruled by the OF definitions but conntrack
        # status can change the datapath. Let's check the rules in two
        # attempts
        for _ in range(2):
            with utils.StatefulConnection(
                    client_ssh[0], srv_ssh, srv_ip, tcp_port) as con:
                self._test_connection_and_log(con)
            for port in range(tcp_port, tcp_port + 3):
                with utils.StatefulConnection(
                        client_ssh[1], srv_ssh, srv_ip, port) as con:
                    self._test_connection_and_log(con)

    def _test_remove_sec_grp_from_active_vm(self):
        """Tests the following:

        1. Create SG associated with ICMP rule
        2. Create Port (assoiated to SG #1) and use it to create the VM
        3. Ping the VM, expected should be PASS
        4. Remove the security group from VM by Port update
        5. Ping the VM, expected should be FAIL
        """
        secgrp = self._create_security_group('test_sg')
        self.security_groups.append(secgrp)
        self.create_pingable_secgroup_rule(secgrp['id'])

        ex_port = self.create_port(
            self.network, fixed_ips=[{'subnet_id': self.subnet['id']}],
            security_groups=[secgrp['id']])
        fip = self.create_vm_testing_sec_grp(
            num_servers=1, security_groups=[{'name': secgrp['name']}],
            ports=[ex_port])[1][0]

        self.ping_ip_address(fip['floating_ip_address'])
        self.client.update_port(ex_port['id'],
                                security_groups=[])
        self.ping_ip_address(fip['floating_ip_address'],
                             should_succeed=False)


class StatefulNetworkSecGroupTest(BaseNetworkSecGroupTest):
    stateless_sg = False
    ipv6_mode = None

    @decorators.idempotent_id('3d73ec1a-2ec6-45a9-b0f8-04a283d9d764')
    def test_default_sec_grp_scenarios(self):
        self._test_default_sec_grp_scenarios()

    @decorators.idempotent_id('3d73ec1a-2ec6-45a9-b0f8-04a283d9d864')
    def test_protocol_number_rule(self):
        self._test_protocol_number_rule()

    @decorators.idempotent_id('3d73ec1a-2ec6-45a9-b0f8-04a283d9d964')
    def test_two_sec_groups(self):
        self._test_two_sec_groups()

    @decorators.idempotent_id('3d73ec1a-2ec6-45a9-b0f8-04a283d9d664')
    def test_ip_prefix(self):
        cidr = self.subnet['cidr']
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_ip_prefix': cidr}]
        self._test_ip_prefix(rule_list, should_succeed=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a01cd2ef-3cfc-4614-8aac-9d1333ea21dd')
    def test_ip_prefix_negative(self):
        # define bad CIDR
        cidr = '10.100.0.254/32'
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_ip_prefix': cidr}]
        self._test_ip_prefix(rule_list, should_succeed=False)

    @decorators.idempotent_id('7ed39b86-006d-40fb-887a-ae46693dabc9')
    def test_remote_group(self):
        self._test_remote_group()

    @testtools.skipUnless(
        CONF.neutron_plugin_options.firewall_driver in ['openvswitch', 'ovn'],
        "Openvswitch agent or Ml2/OVN is required to run this test")
    @decorators.idempotent_id('678dd4c0-2953-4626-b89c-8e7e4110ec4b')
    @tempest_utils.requires_ext(extension="address-group", service="network")
    @tempest_utils.requires_ext(
        extension="security-groups-remote-address-group", service="network")
    def test_remote_group_and_remote_address_group(self):
        self._test_remote_group_and_remote_address_group()

    @decorators.idempotent_id('f07d0159-8f9e-4faa-87f5-a869ab0ad488')
    def test_multiple_ports_secgroup_inheritance(self):
        self._test_multiple_ports_secgroup_inheritance()

    @decorators.idempotent_id('f07d0159-8f9e-4faa-87f5-a869ab0ad489')
    def test_multiple_ports_portrange_remote(self):
        self._test_multiple_ports_portrange_remote()

    @decorators.idempotent_id('f07d0159-8f9e-4faa-87f5-a869ab0ad490')
    def test_intra_sg_isolation(self):
        """Test intra security group isolation

        This test creates a security group that does not allow ingress
        packets from vms of the same security group. The purpose of this
        test is to verify that intra SG traffic is properly blocked, while
        traffic like metadata and DHCP remains working due to the
        allow-related behavior of the egress rules (added via default).
        """
        # create a security group and make it loginable
        secgrp = self._create_security_group('secgrp')

        # remove all rules and add ICMP, DHCP and metadata as egress,
        # and ssh as ingress.
        for sgr in secgrp['security_group_rules']:
            self.client.delete_security_group_rule(sgr['id'])

        self.create_loginable_secgroup_rule(secgroup_id=secgrp['id'])
        rule_list = [{'direction': constants.EGRESS_DIRECTION,
                      'protocol': constants.PROTO_NAME_TCP,
                      'remote_ip_prefix': '169.254.169.254/32',
                      'description': 'metadata out',
                      },
                     {'direction': constants.EGRESS_DIRECTION,
                      'protocol': constants.PROTO_NAME_UDP,
                      'port_range_min': '67',
                      'port_range_max': '67',
                      'description': 'dhcpv4 out',
                      },
                     {'direction': constants.EGRESS_DIRECTION,
                      'protocol': constants.PROTO_NAME_ICMP,
                      'description': 'ping out',
                      },
                     ]
        self.create_secgroup_rules(rule_list, secgroup_id=secgrp['id'])

        # go vms, go!
        ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            num_servers=2,
            security_groups=[{'name': secgrp['name']}])

        # verify SSH functionality. This will ensure that servers were
        # able to reach dhcp + metadata servers
        for fip in fips:
            self.check_connectivity(fip['floating_ip_address'],
                                    CONF.validation.image_ssh_user,
                                    self.keypair['private_key'])

        # try to ping instances without intra SG permission (should fail)
        self.check_remote_connectivity(
            ssh_clients[0], fips[1]['fixed_ip_address'],
            should_succeed=False)
        self.check_remote_connectivity(
            ssh_clients[1], fips[0]['fixed_ip_address'],
            should_succeed=False)

        # add intra sg rule. This will allow packets from servers that
        # are in the same sg
        rule_list = [{'direction': constants.INGRESS_DIRECTION,
                      'remote_group_id': secgrp['id']}]
        self.create_secgroup_rules(rule_list, secgroup_id=secgrp['id'])

        # try to ping instances with intra SG permission
        self.check_remote_connectivity(
            ssh_clients[0], fips[1]['fixed_ip_address'])
        self.check_remote_connectivity(
            ssh_clients[1], fips[0]['fixed_ip_address'])

    @decorators.idempotent_id('cd66b826-d86c-4fb4-ab37-17c8391753cb')
    def test_overlapping_sec_grp_rules(self):
        self._test_overlapping_sec_grp_rules()

    @decorators.idempotent_id('96dcd5ff-9d45-4e0d-bea0-0b438cbd388f')
    def test_remove_sec_grp_from_active_vm(self):
        self._test_remove_sec_grp_from_active_vm()

    @decorators.idempotent_id('01f0ddca-b049-47eb-befd-82acb502c9ec')
    def test_established_tcp_session_after_re_attachinging_sg(self):
        """Test existing connection remain open after sg has been re-attached

        Verifies that new packets can pass over the existing connection when
        the security group has been removed from the server and then added
        back
        """

        ssh_sg = self._create_security_group('ssh_sg')
        self.create_loginable_secgroup_rule(secgroup_id=ssh_sg['id'])
        vm_ssh, fips, vms = self.create_vm_testing_sec_grp(
                security_groups=[{'name': ssh_sg['name']}])
        sg = self._create_security_group('sg')
        nc_rule = [{'protocol': constants.PROTO_NUM_TCP,
                    'direction': constants.INGRESS_DIRECTION,
                    'port_range_min': 6666,
                    'port_range_max': 6666}]
        self.create_secgroup_rules(nc_rule, secgroup_id=sg['id'])
        srv_port = self.client.list_ports(network_id=self.network['id'],
                device_id=vms[1]['server']['id'])['ports'][0]
        srv_ip = srv_port['fixed_ips'][0]['ip_address']
        with utils.StatefulConnection(
                vm_ssh[0], vm_ssh[1], srv_ip, 6666) as con:
            self.client.update_port(srv_port['id'],
                    security_groups=[ssh_sg['id'], sg['id']])
            self._test_connection_and_log(con)
        with utils.StatefulConnection(
                vm_ssh[0], vm_ssh[1], srv_ip, 6666) as con:
            self.client.update_port(
                    srv_port['id'], security_groups=[ssh_sg['id']])
            self._test_connection_and_log(con, should_pass=False)
        with utils.StatefulConnection(
                vm_ssh[0], vm_ssh[1], srv_ip, 6666) as con:
            self.client.update_port(srv_port['id'],
                    security_groups=[ssh_sg['id'], sg['id']])
            self._test_connection_and_log(con)
            self.client.update_port(srv_port['id'],
                    security_groups=[ssh_sg['id']])
            self._test_connection_and_log(con, should_pass=False)
            self.client.update_port(srv_port['id'],
                    security_groups=[ssh_sg['id'], sg['id']])
            self._test_connection_and_log(con)

    @decorators.idempotent_id('4a724164-bbc0-4029-a844-644ece66c026')
    def test_connectivity_between_vms_using_different_sec_groups(self):
        self._test_connectivity_between_vms_using_different_sec_groups()


@testtools.skipIf(
    CONF.neutron_plugin_options.firewall_driver in ['openvswitch', 'None'],
    "Firewall driver other than 'openvswitch' is required to use "
    "stateless security groups.")
class StatelessNetworkSecGroupIPv4Test(BaseNetworkSecGroupTest):
    required_extensions = ['security-group', 'stateful-security-group']
    stateless_sg = True
    ipv6_mode = None

    @decorators.idempotent_id('9e193e3f-56f2-4f4e-886c-988a147958ef')
    def test_default_sec_grp_scenarios(self):
        self._test_default_sec_grp_scenarios()

    @decorators.idempotent_id('afae8654-a389-4887-b21d-7f07ec350177')
    def test_protocol_number_rule(self):
        self._test_protocol_number_rule()

    @decorators.idempotent_id('b51cc0eb-8f9a-49e7-96ab-61cd31243b67')
    def test_two_sec_groups(self):
        self._test_two_sec_groups()

    @decorators.idempotent_id('07985496-58da-4c1f-a6ef-2fdd88128a81')
    def test_ip_prefix(self):
        cidr = self.subnet['cidr']
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_ip_prefix': cidr}]
        self._test_ip_prefix(rule_list, should_succeed=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('1ad469c4-0d8f-42ae-8ec3-46cc424565c4')
    def test_ip_prefix_negative(self):
        # define bad CIDR
        cidr = '10.100.0.254/32'
        rule_list = [{'protocol': constants.PROTO_NUM_ICMP,
                      'direction': constants.INGRESS_DIRECTION,
                      'remote_ip_prefix': cidr}]
        self._test_ip_prefix(rule_list, should_succeed=False)

    @decorators.idempotent_id('fa1e93bf-67c5-4590-9962-38ee1f43a46a')
    def test_remote_group(self):
        self._test_remote_group()

    @testtools.skipUnless(
        CONF.neutron_plugin_options.firewall_driver in ['openvswitch', 'ovn'],
        "Openvswitch agent or Ml2/OVN is required to run this test")
    @decorators.idempotent_id('9fae530d-2711-4c61-a4a5-8efe6e58ab14')
    @tempest_utils.requires_ext(extension="address-group", service="network")
    @tempest_utils.requires_ext(
        extension="security-groups-remote-address-group", service="network")
    def test_remote_group_and_remote_address_group(self):
        self._test_remote_group_and_remote_address_group()

    @decorators.idempotent_id('4f1eb6db-ae7f-4f26-b371-cbd8363f9b0b')
    def test_multiple_ports_secgroup_inheritance(self):
        self._test_multiple_ports_secgroup_inheritance()

    @decorators.idempotent_id('4043ca0a-eabb-4198-be53-3d3051cc0804')
    def test_multiple_ports_portrange_remote(self):
        self._test_multiple_ports_portrange_remote()

    @decorators.idempotent_id('bfe25138-ceac-4944-849a-b9b90aff100f')
    def test_overlapping_sec_grp_rules(self):
        self._test_overlapping_sec_grp_rules()

    @decorators.idempotent_id('e4340e47-39cd-49ed-967c-fc2c40b47c5a')
    def test_remove_sec_grp_from_active_vm(self):
        self._test_remove_sec_grp_from_active_vm()

    @decorators.idempotent_id('8d4753cc-cd7a-48a0-8ece-e11efce2af10')
    def test_reattach_sg_with_changed_mode(self):
        sg_kwargs = {'stateful': True}
        secgrp = self.os_primary.network_client.create_security_group(
            name=data_utils.rand_name('secgrp'), **sg_kwargs)['security_group']
        # add cleanup
        self.security_groups.append(secgrp)

        # now configure sec group to support required connectivity
        self.create_pingable_secgroup_rule(secgroup_id=secgrp['id'])
        # and create server
        ssh_clients, fips, servers = self.create_vm_testing_sec_grp(
            num_servers=1, security_groups=[{'name': secgrp['name']}])
        server_ports = self.network_client.list_ports(
            device_id=servers[0]['server']['id'])['ports']

        # make sure connectivity works
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=True)
        # remove SG from ports
        for port in server_ports:
            self.network_client.update_port(port['id'], security_groups=[])
        # make sure there is now no connectivity as there's no SG attached
        # to the port
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=False)

        # Update SG to be stateless
        self.os_primary.network_client.update_security_group(
            secgrp['id'], stateful=False)
        # Add SG back to the ports
        for port in server_ports:
            self.network_client.update_port(
                port['id'], security_groups=[secgrp['id']])
        # Make sure connectivity works fine again
        self.ping_ip_address(fips[0]['floating_ip_address'],
                             should_succeed=True)

    @decorators.idempotent_id('7ede9ab5-a615-46c5-9dea-cf2aa1ea43cb')
    def test_connectivity_between_vms_using_different_sec_groups(self):
        self._test_connectivity_between_vms_using_different_sec_groups()

    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @decorators.idempotent_id('c3bb8073-97a2-4bea-a6fb-0a9d2e4df13f')
    def test_packets_of_any_connection_state_can_reach_dest(self):
        TEST_TCP_PORT = 1022
        PKT_TYPES = [
            {'nping': 'syn', 'tcpdump': 'tcp-syn'},
            {'nping': 'ack', 'tcpdump': 'tcp-ack'},
            {'nping': 'syn,ack', 'tcpdump': 'tcp-syn|tcp-ack'},
            {'nping': 'rst', 'tcpdump': 'tcp-rst'},
            {'nping': 'fin', 'tcpdump': 'tcp-fin'},
            {'nping': 'psh', 'tcpdump': 'tcp-push'}]
        ssh_clients, fips, servers, _ = self._create_client_and_server_vms(
            TEST_TCP_PORT, use_advanced_image=True)

        self._check_cmd_installed_on_server(
            ssh_clients['server'], servers['server']['server'], 'nping')
        self._check_cmd_installed_on_server(
            ssh_clients['client'], servers['client']['server'], 'tcpdump')
        server_port = self.network_client.show_port(
            fips['server']['port_id'])['port']
        server_ip_command = ip.IPCommand(ssh_client=ssh_clients['server'])
        addresses = server_ip_command.list_addresses(port=server_port)
        port_iface = ip.get_port_device_name(addresses, server_port)

        def _get_file_suffix(pkt_type):
            return pkt_type['tcpdump'].replace(
                'tcp-', '').replace('|', '')

        for pkt_type in PKT_TYPES:
            file_suffix = _get_file_suffix(pkt_type)
            capture_script_path = "/tmp/capture_%s.sh" % file_suffix
            capture_out = "/tmp/capture_%s.out" % file_suffix
            capture_script = get_capture_script(
                port_iface, TEST_TCP_PORT, pkt_type['tcpdump'], capture_out)
            ssh_clients['server'].execute_script(
                'echo \'%s\' > %s' % (capture_script, capture_script_path))
            ssh_clients['server'].execute_script(
                "bash %s" % capture_script_path, become_root=True)

        for pkt_type in PKT_TYPES:
            ssh_clients['client'].execute_script(
                "nping --tcp -p %(tcp_port)s --flags %(tcp_flag)s --ttl 10 "
                "%(ip_address)s -c 3" % {
                    'tcp_port': TEST_TCP_PORT,
                    'tcp_flag': pkt_type['nping'],
                    'ip_address': fips['server']['fixed_ip_address']},
                become_root=True)

        def _packtet_received(pkt_type):
            file_suffix = _get_file_suffix(pkt_type)
            expected_msg = "1 packet captured"
            result = ssh_clients['server'].execute_script(
                "cat {path} || echo '{path} not exists yet'".format(
                    path="/tmp/capture_%s.out" % file_suffix))
            return expected_msg in result

        for pkt_type in PKT_TYPES:
            utils.wait_until_true(
                lambda: _packtet_received(pkt_type),
                timeout=10,
                exception=RuntimeError(
                    'No TCP packet of type %s received by server %s' % (
                        pkt_type['nping'],
                        fips['server']['fixed_ip_address'])))

    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @decorators.idempotent_id('14c4af2c-8077-4756-a6e3-6bebd642ed92')
    def test_fragmented_traffic_is_accepted(self):
        ssh_clients, fips, servers, security_groups = (
            self._create_client_and_server_vms(use_advanced_image=True))
        if CONF.neutron_plugin_options.default_image_is_advanced:
            username = CONF.validation.image_ssh_user
        else:
            username = CONF.neutron_plugin_options.advanced_image_ssh_user

        # make sure tcp connectivity to vms works fine
        for fip in fips.values():
            self.check_connectivity(
                fip['floating_ip_address'],
                username,
                self.keypair['private_key'])

        # Check that ICMP packets bigger than MTU aren't working without
        # fragmentation allowed
        self.check_remote_connectivity(
            ssh_clients['client'], fips['server']['fixed_ip_address'],
            mtu=self.network['mtu'] + 1, fragmentation=False,
            should_succeed=False)
        # and are working fine with fragmentation enabled:
        self.check_remote_connectivity(
            ssh_clients['client'], fips['server']['fixed_ip_address'],
            mtu=self.network['mtu'] + 1, fragmentation=True,
            should_succeed=True)


class StatelessSecGroupDualStackBase(BaseNetworkSecGroupTest):
    required_extensions = ['security-group', 'stateful-security-group']
    stateless_sg = True

    def _get_port_cidrs(self, port):
        ips = []
        subnet_cidrs = {}
        for fixed_ip in port['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            subnet_cidr = subnet_cidrs.get('subnet_id')
            if not subnet_cidr:
                subnet = self.client.show_subnet(subnet_id)['subnet']
                subnet_cidr = netaddr.IPNetwork(subnet['cidr'])
                subnet_cidrs[subnet_id] = subnet_cidr
            ips.append(
                netaddr.IPNetwork(
                    "%s/%s" % (fixed_ip['ip_address'], subnet_cidr.prefixlen)))
        LOG.debug("On port %s found IP cidrs: %s", port['id'], ips)
        return ips

    def _test_default_sec_grp_scenarios(self):
        # Make "regular" test like for IPv4 case
        server_ssh_clients, _, servers = (
            super()._test_default_sec_grp_scenarios())

        # And additionally ensure that IPv6 addresses are configured properly
        # in the VM
        for ssh_client, server in zip(server_ssh_clients, servers):
            ip_cmd = ip.IPCommand(ssh_client=ssh_client)
            ports = self.client.list_ports(
                device_id=server['server']['id'])['ports']
            for port in ports:
                configured_cidrs = [ip.network for ip in
                                    ip_cmd.list_addresses(port=port)]
                for port_cidr in self._get_port_cidrs(port):
                    self.assertIn(port_cidr, configured_cidrs)


@testtools.skipIf(
    CONF.neutron_plugin_options.firewall_driver in ['openvswitch', 'None'],
    "Firewall driver other than 'openvswitch' is required to use "
    "stateless security groups.")
class StatelessSecGroupDualStackSlaacTest(StatelessSecGroupDualStackBase):
    ipv6_mode = 'slaac'

    @decorators.idempotent_id('e7d64384-ea6a-40aa-b454-854f0990153c')
    def test_default_sec_grp_scenarios(self):
        self._test_default_sec_grp_scenarios()


@testtools.skipIf(
    CONF.neutron_plugin_options.firewall_driver in ['openvswitch', 'None'],
    "Firewall driver other than 'openvswitch' is required to use "
    "stateless security groups.")
class StatelessSecGroupDualStackDHCPv6StatelessTest(
        StatelessSecGroupDualStackBase):
    ipv6_mode = 'dhcpv6-stateless'

    @decorators.idempotent_id('c61c127c-e08f-4ddf-87a3-58b3c86e5476')
    def test_default_sec_grp_scenarios(self):
        self._test_default_sec_grp_scenarios()



# Copyright 2016
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.fwaas.api import fwaas_v2_base as v2_base


CONF = config.CONF
DEFAULT_FWG = 'default'


class FWaaSv2ExtensionTestJSON(v2_base.BaseFWaaSTest):

    """List of tests

    Tests the following operations in the Neutron API using the REST client
    for Neutron:

        List firewall rules
        Create firewall rule
        Update firewall rule
        Delete firewall rule
        Show firewall rule
        List firewall policies
        Create firewall policy
        Update firewall policy
        Insert firewall rule to policy
        Remove firewall rule from policy
        Insert firewall rule after/before rule in policy
        Update firewall policy audited attribute
        Delete firewall policy
        Show firewall policy
        List firewall group
        Create firewall group
        Update firewall group
        Delete firewall group
        Show firewall group
    """

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        if not utils.is_extension_enabled('fwaas_v2', 'network'):
            msg = "FWaaS v2  Extension not enabled."
            raise cls.skipException(msg)

    def setUp(self):
        super().setUp()
        self.fw_rule_1 = self.create_firewall_rule(action="allow",
                                                 protocol="tcp")
        self.fw_rule_2 = self.create_firewall_rule(action="deny",
                                                 protocol="udp")
        self.fw_policy_1 = self.create_firewall_policy(
            firewall_rules=[self.fw_rule_1['id']])
        self.fw_policy_2 = self.create_firewall_policy(
            firewall_rules=[self.fw_rule_2['id']])

    def _create_router_interfaces(self):
        network_1 = self.create_network()
        network_2 = self.create_network()

        cidr = netaddr.IPNetwork(CONF.network.project_network_cidr)
        mask_bits = CONF.network.project_network_mask_bits

        subnet_cidr_1 = list(cidr.subnet(mask_bits))[-1]
        subnet_cidr_2 = list(cidr.subnet(mask_bits))[-2]
        subnet_1 = self.create_subnet(network_1, cidr=subnet_cidr_1,
            mask_bits=mask_bits)
        subnet_2 = self.create_subnet(network_2, cidr=subnet_cidr_2,
            mask_bits=mask_bits)

        router = self.create_router(
            data_utils.rand_name('router-'),
            admin_state_up=True)
        self.addCleanup(self._try_delete_router, router)

        intf_1 = self.routers_client.add_router_interface(router['id'],
            subnet_id=subnet_1['id'])
        intf_2 = self.routers_client.add_router_interface(router['id'],
            subnet_id=subnet_2['id'])

        return intf_1, intf_2

    def _try_delete_router(self, router):
        # delete router, if it exists
        try:
            self.delete_router(router)
        # if router is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

    def _try_delete_policy(self, policy_id):
        # delete policy, if it exists
        try:
            self.firewall_policies_client.delete_firewall_policy(policy_id)
        # if policy is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

    def _try_delete_rule(self, rule_id):
        # delete rule, if it exists
        try:
            self.firewall_rules_client.delete_firewall_rule(rule_id)
        # if rule is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

    def _try_delete_firewall_group(self, fwg_id):
        # delete firewall group, if it exists
        try:
            self.firewall_groups_client.delete_firewall_group(fwg_id)
        # if firewall group is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

        self.firewall_groups_client.wait_for_resource_deletion(fwg_id)

    def _wait_until_ready(self, fwg_id):
        target_states = ('ACTIVE', 'CREATED')

        def _wait():
            firewall_group = self.firewall_groups_client.show_firewall_group(
                fwg_id)
            firewall_group = firewall_group['firewall_group']
            return firewall_group['status'] in target_states

        if not test_utils.call_until_true(_wait, CONF.network.build_timeout,
                                          CONF.network.build_interval):
            m = ("Timed out waiting for firewall_group %s to reach %s "
                 "state(s)" %
                 (fwg_id, target_states))
            raise lib_exc.TimeoutException(m)

    def _wait_until_deleted(self, fwg_id):
        def _wait():
            try:
                fwg = self.firewall_groups_client.show_firewall_group(fwg_id)
            except lib_exc.NotFound:
                return True

            fwg_status = fwg['firewall_group']['status']
            if fwg_status == 'ERROR':
                raise lib_exc.DeleteErrorException(resource_id=fwg_id)

        if not test_utils.call_until_true(_wait, CONF.network.build_timeout,
                                          CONF.network.build_interval):
            m = ("Timed out waiting for firewall_group %s deleted" % fwg_id)
            raise lib_exc.TimeoutException(m)

    @decorators.idempotent_id('ddccfa87-4af7-48a6-9e50-0bd0ad1348cb')
    def test_list_firewall_rules(self):
        # List firewall rules
        fw_rules = self.firewall_rules_client.list_firewall_rules()
        fw_rules = fw_rules['firewall_rules']
        self.assertIn((self.fw_rule_1['id'],
                       self.fw_rule_1['name'],
                       self.fw_rule_1['action'],
                       self.fw_rule_1['protocol'],
                       self.fw_rule_1['ip_version'],
                       self.fw_rule_1['enabled']),
                      [(m['id'],
                        m['name'],
                        m['action'],
                        m['protocol'],
                        m['ip_version'],
                        m['enabled']) for m in fw_rules])

    @decorators.idempotent_id('ffc009fa-cd17-4029-8025-c4b81a7dd923')
    def test_create_update_delete_firewall_rule(self):
        # Create firewall rule
        body = self.firewall_rules_client.create_firewall_rule(
            name=data_utils.rand_name("fw-rule"),
            action="allow",
            protocol="tcp")
        fw_rule_id = body['firewall_rule']['id']
        self.addCleanup(self._try_delete_rule, fw_rule_id)

        # Update firewall rule
        body = self.firewall_rules_client.update_firewall_rule(fw_rule_id,
                                                               action="deny")
        self.assertEqual("deny", body["firewall_rule"]['action'])

        # Delete firewall rule
        self.firewall_rules_client.delete_firewall_rule(fw_rule_id)
        # Confirm deletion
        fw_rules = self.firewall_rules_client.list_firewall_rules()
        self.assertNotIn(fw_rule_id,
                         [m['id'] for m in fw_rules['firewall_rules']])

    @decorators.idempotent_id('76b07afc-444e-4bb9-abec-9b8c5f994dcd')
    def test_show_firewall_rule(self):
        # show a created firewall rule
        fw_rule = self.firewall_rules_client.show_firewall_rule(
            self.fw_rule_1['id'])
        for key, value in fw_rule['firewall_rule'].items():
            if key != 'firewall_policy_id':
                self.assertEqual(self.fw_rule_1[key], value)
            # This check is placed because we cannot modify policy during
            # Create/Update of Firewall Rule but we can see the association
            # of a Firewall Rule with the policies it belongs to.

    @decorators.idempotent_id('f6b83902-746f-4e74-9403-2ec9899583a3')
    def test_list_firewall_policies(self):
        fw_policies = self.firewall_policies_client.list_firewall_policies()
        fw_policies = fw_policies['firewall_policies']
        self.assertIn((self.fw_policy_1['id'],
                       self.fw_policy_1['name'],
                       self.fw_policy_1['firewall_rules']),
                      [(m['id'],
                        m['name'],
                        m['firewall_rules']) for m in fw_policies])

    @decorators.idempotent_id('6ef9bd02-7349-4d61-8d1f-80479f64d904')
    def test_create_update_delete_firewall_policy(self):
        # Create firewall policy
        body = self.firewall_policies_client.create_firewall_policy(
            name=data_utils.rand_name("fw-policy"))
        fw_policy_id = body['firewall_policy']['id']
        self.addCleanup(self._try_delete_policy, fw_policy_id)

        # Update firewall policy
        body = self.firewall_policies_client.update_firewall_policy(
            fw_policy_id,
            name="updated_policy")
        updated_fw_policy = body["firewall_policy"]
        self.assertEqual("updated_policy", updated_fw_policy['name'])

        # Delete firewall policy
        self.firewall_policies_client.delete_firewall_policy(fw_policy_id)
        # Confirm deletion
        fw_policies = self.firewall_policies_client.list_firewall_policies()
        fw_policies = fw_policies['firewall_policies']
        self.assertNotIn(fw_policy_id, [m['id'] for m in fw_policies])

    @decorators.idempotent_id('164381de-61f4-483f-9a5a-48105b8e70e2')
    def test_show_firewall_policy(self):
        # show a created firewall policy
        fw_policy = self.firewall_policies_client.show_firewall_policy(
            self.fw_policy_1['id'])
        fw_policy = fw_policy['firewall_policy']
        for key, value in fw_policy.items():
            self.assertEqual(self.fw_policy_1[key], value)

    @decorators.idempotent_id('48dfcd75-3924-479d-bb65-b3ed33397663')
    def test_create_show_delete_firewall_group(self):
        # create router and add interfaces
        intf_1, intf_2 = self._create_router_interfaces()

        # Create firewall_group
        body = self.firewall_groups_client.create_firewall_group(
            name=data_utils.rand_name("firewall_group"),
            ingress_firewall_policy_id=self.fw_policy_1['id'],
            egress_firewall_policy_id=self.fw_policy_2['id'],
            ports=[intf_1['port_id'], intf_2['port_id']])
        created_firewall_group = body['firewall_group']
        fwg_id = created_firewall_group['id']

        # Wait for the firewall resource to become ready
        self._wait_until_ready(fwg_id)

        # show created firewall_group
        firewall_group = self.firewall_groups_client.show_firewall_group(
            fwg_id)
        fwg = firewall_group['firewall_group']

        for key, value in fwg.items():
            if key == 'status':
                continue
            self.assertEqual(created_firewall_group[key], value)

        # list firewall_groups
        firewall_groups = self.firewall_groups_client.list_firewall_groups()
        fwgs = firewall_groups['firewall_groups']
        self.assertIn((created_firewall_group['id'],
                       created_firewall_group['name'],
                       created_firewall_group['ingress_firewall_policy_id'],
                       created_firewall_group['egress_firewall_policy_id']),
                      [(m['id'],
                        m['name'],
                        m['ingress_firewall_policy_id'],
                        m['egress_firewall_policy_id']) for m in fwgs])

        # Disassociate all port with this firewall group
        self.firewall_groups_client.update_firewall_group(fwg_id, ports=[])
        # Delete firewall_group
        self.firewall_groups_client.delete_firewall_group(fwg_id)

        # Wait for the firewall group to be deleted
        self._wait_until_deleted(fwg_id)

        # Confirm deletion
        firewall_groups = self.firewall_groups_client.list_firewall_groups()
        fwgs = firewall_groups['firewall_groups']
        self.assertNotIn(fwg_id, [m['id'] for m in fwgs])

    @decorators.idempotent_id('e021baab-d4f7-4bad-b382-bde4946e0e0b')
    def test_update_firewall_group(self):
        # create router and add interfaces
        intf_1, intf_2 = self._create_router_interfaces()

        # Create firewall_group
        body = self.firewall_groups_client.create_firewall_group(
            name=data_utils.rand_name("firewall_group"),
            ingress_firewall_policy_id=self.fw_policy_1['id'],
            egress_firewall_policy_id=self.fw_policy_2['id'],
            ports=[intf_1['port_id']])
        created_firewall_group = body['firewall_group']
        fwg_id = created_firewall_group['id']
        self.addCleanup(self._try_delete_firewall_group, fwg_id)

        # Wait for the firewall resource to become ready
        self._wait_until_ready(fwg_id)

        # Update firewall group
        body = self.firewall_groups_client.update_firewall_group(
            fwg_id,
            ports=[intf_2['port_id']])
        updated_fwg = body["firewall_group"]
        self.assertEqual([intf_2['port_id']], updated_fwg['ports'])
        # Wait for the firewall resource to become ready
        self._wait_until_ready(fwg_id)
        # Disassociate all ports with this firewall group
        self.firewall_groups_client.update_firewall_group(fwg_id, ports=[])
        # Delete firewall_group
        self.firewall_groups_client.delete_firewall_group(fwg_id)

    @decorators.idempotent_id('a1f524d8-5336-4769-aa7b-0830bb4df6c8')
    def test_error_on_create_firewall_group_name_default(self):
        try:
            # Create firewall_group with name == reserved default fwg
            self.firewall_groups_client.create_firewall_group(
                name=DEFAULT_FWG)
        except lib_exc.Conflict:
            pass

    @decorators.idempotent_id('fd24db16-c8cb-4cb4-ba60-b0cd18a66b83')
    def test_default_fwg_created_on_list_firewall_groups(self):
        fw_groups = self.firewall_groups_client.list_firewall_groups()
        fw_groups = fw_groups['firewall_groups']
        self.assertIn(DEFAULT_FWG,
                      [g['name'] for g in fw_groups])



# Copyright (c) 2018 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import re
import subprocess

import netaddr
from neutron_lib._i18n import _
from neutron_lib import constants
from oslo_log import log
from oslo_utils import excutils
from tempest.common import waiters

from neutron_tempest_plugin.common import shell
from neutron_tempest_plugin.common import utils as common_utils


LOG = log.getLogger(__name__)


class IPCommand:

    sudo = 'sudo'
    ip_path = '/sbin/ip'

    def __init__(self, ssh_client=None, timeout=None, namespace=None):
        self.ssh_client = ssh_client
        self.timeout = timeout
        self.namespace = namespace

    def get_command(self, obj, *command):
        command_line = '{sudo!s} {ip_path!r} '.format(sudo=self.sudo,
                                                     ip_path=self.ip_path)
        if self.namespace:
            command_line += 'netns exec {ns_name!s} {ip_path!r} '.format(
                ns_name=self.namespace, ip_path=self.ip_path)
        command_line += '{object!s} {command!s}'.format(
            object=obj,
            command=subprocess.list2cmdline([str(c) for c in command]))
        return command_line

    def execute(self, obj, *command):
        command_line = self.get_command(obj, *command)
        return shell.execute(command_line, ssh_client=self.ssh_client,
                             timeout=self.timeout).stdout

    def configure_vlan(self, addresses, port, vlan_tag, subport_ips, mac=None):
        port_device = get_port_device_name(addresses=addresses, port=port)
        subport_device = '{!s}.{!s}'.format(port_device, vlan_tag)
        LOG.debug('Configuring VLAN subport interface %r on top of interface '
                  '%r with IPs: %s', subport_device, port_device,
                  ', '.join(subport_ips))

        self.add_link(link=port_device, name=subport_device, link_type='vlan',
                      segmentation_id=vlan_tag)
        if mac:
            self.set_link_address(address=mac, device=subport_device)
        self.set_link(device=subport_device, state='up')
        for subport_ip in subport_ips:
            self.add_address(address=subport_ip, device=subport_device)
        return subport_device

    def configure_vlan_subport(self, port, subport, vlan_tag, subnets):
        addresses = self.list_addresses()
        try:
            subport_device = get_port_device_name(addresses=addresses,
                                                  port=subport)
        except ValueError:
            pass
        else:
            LOG.debug('Interface %r already configured.', subport_device)
            return subport_device

        subport_ips = [
            "{!s}/{!s}".format(ip, prefix_len)
            for ip, prefix_len in _get_ip_address_prefix_len_pairs(
                port=subport, subnets=subnets)]
        if not subport_ips:
            raise ValueError(_(
                "Unable to get IP address and subnet prefix lengths for "
                "subport"))

        return self.configure_vlan(addresses, port, vlan_tag, subport_ips,
                                   subport['mac_address'])

    def configure_inner_vlan(self, port, vlan_tag, ip_addresses):
        addresses = self.list_addresses()
        try:
            subport_device = get_vlan_device_name(addresses, ip_addresses)
        except ValueError:
            pass
        else:
            LOG.debug('Interface %r already configured.', subport_device)
            return subport_device

        return self.configure_vlan(addresses, port, vlan_tag, ip_addresses)

    # NOTE(ralonsoh): some projects, like whitebox-neutron-tempest-plugin, are
    # using ``configure_vlan_transparent`` method. The concept of "inner VLAN"
    # does not exist in the VLAN transparency feature.
    configure_vlan_transparent = configure_inner_vlan

    def list_namespaces(self):
        namespaces_output = self.execute("netns")
        ns_list = []
        for ns_line in namespaces_output.split("\n"):
            ns_list.append(ns_line.split(" ", 1)[0])
        return ns_list

    def list_addresses(self, device=None, ip_addresses=None, port=None,
                       subnets=None):
        command = ['list']
        if device:
            command += ['dev', device]
        output = self.execute('address', *command)
        addresses = list(parse_addresses(output))

        return list_ip_addresses(addresses=addresses,
                                 ip_addresses=ip_addresses, port=port,
                                 subnets=subnets)

    def add_link(self, name, link_type, link=None, segmentation_id=None):
        command = ['add']
        if link:
            command += ['link', link]
        command += ['name', name, 'type', link_type]
        if segmentation_id:
            command += ['id', segmentation_id]
        return self.execute('link', *command)

    def set_link_address(self, address, device):
        command = ['set', 'address', address, 'dev', device]
        return self.execute('link', *command)

    def set_link(self, device, state=None):
        command = ['set', 'dev', device]
        if state:
            command.append(state)
        return self.execute('link', *command)

    def add_address(self, address, device):
        # ip addr add 192.168.1.1/24 dev em1
        return self.execute('address', 'add', address, 'dev', device)

    def delete_address(self, address, device):
        # ip addr del 192.168.1.1/24 dev em1
        return self.execute('address', 'del', address, 'dev', device)

    def add_route(self, address, device, gateway=None, ip_version=4):
        if gateway:
            return self.execute(
                'route', 'add', address, 'via', gateway, 'dev', device)
        else:
            return self.execute(
                f'-{ip_version}', 'route', 'add', address, 'dev', device)

    def delete_route(self, address, device, ip_version=4):
        return self.execute(
            f'-{ip_version}', 'route', 'del', address, 'dev', device)

    def list_routes(self, *args, device=None, ip_version=4):
        if not args and device:
            args = ("dev", device)
        output = self.execute(f'-{ip_version}', 'route', 'show', *args)
        return list(parse_routes(output))

    def get_nic_name_by_mac(self, mac_address):
        nics = self.execute("-o", "link")
        for nic_line in nics.split("\n"):
            if mac_address in nic_line:
                return nic_line.split(":")[1].strip()

    def has_dadfailed(self, device):
        """Check if device has any IPv6 addresses in dadfailed state"""
        output = self.execute('address', 'show', 'dev', device)
        return 'dadfailed' in output


def parse_addresses(command_output):
    address = device = None
    addresses = []
    for i, line in enumerate(command_output.split('\n')):
        try:
            line_number = i + 1
            fields = line.strip().split()
            if not fields:
                continue
            indent = line.index(fields[0] + ' ')
            if indent == 0:
                # example of line
                # 2: enp0s25: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc pfifo_fast state UP qlen 1000  # noqa
                address = None
                name = fields[1]
                if name.endswith(':'):
                    name = name[:-1]
                if '@' in name:
                    name, parent = name.split('@', 1)
                else:
                    parent = None

                if len(fields) > 2:
                    # flags example: <LOOPBACK,UP,LOWER_UP>
                    flags = fields[2]
                    if flags.startswith('<'):
                        flags = flags[1:]
                    if flags.startswith('>'):
                        flags = flags[:-1]
                    flags = flags.split(',')

                device = Device(name=name, parent=parent, flags=flags,
                                properties=dict(parse_properties(fields[3:])))
                LOG.debug("Device parsed: %r", device)

            elif indent == 4:
                address = Address.create(
                    family=fields[0], address=fields[1], device=device,
                    properties=dict(parse_properties(fields[2:])))
                addresses.append(address)
                LOG.debug("Address parsed: %r", address)

            elif indent == 7:
                address.properties.update(parse_properties(fields))
                LOG.debug("Address properties parsed: %r", address.properties)

            else:
                LOG.error("Invalid line indentation: %s", indent)
                raise

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error parsing ip command output at line %d:\n"
                              "%r\n",
                              line_number, line)
            raise

    return addresses


def parse_properties(fields):
    for i, field in enumerate(fields):
        if i % 2 == 0:
            key = field
        else:
            yield key, field


class HasProperties:

    def __getattr__(self, name):
        try:
            return self.properties[name]
        except KeyError:
            pass
        # This should raise AttributeError
        return getattr(super(), name)


class Address(HasProperties,
              collections.namedtuple('Address',
                                     ['family', 'address', 'device',
                                      'properties'])):

    _subclasses = {}

    @classmethod
    def create(cls, family, address, device, properties):
        cls = cls._subclasses.get(family, cls)
        return cls(family=family, address=address, device=device,
                   properties=properties)

    @classmethod
    def register_subclass(cls, family, subclass=None):
        if not issubclass(subclass, cls):
            msg = "{!r} is not sub-class of {!r}".format(cls, Address)
            raise TypeError(msg)
        cls._subclasses[family] = subclass


class Device(HasProperties,
             collections.namedtuple('Device',
                                    ['name', 'parent', 'flags',
                                     'properties'])):
    pass


def register_address_subclass(families):

    def decorator(subclass):
        for family in families:
            Address.register_subclass(family=family, subclass=subclass)
        return subclass

    return decorator


@register_address_subclass(['inet', 'inet6'])
class InetAddress(Address):

    @property
    def ip(self):
        return self.network.ip

    @property
    def network(self):
        return netaddr.IPNetwork(self.address)


def parse_routes(command_output):
    for line in command_output.split('\n'):
        fields = line.strip().split()
        if fields:
            dest = fields[0]
            properties = dict(parse_properties(fields[1:]))
            if dest == 'default':
                dest = constants.IPv4_ANY
                via = properties.get('via')
                if via:
                    dest = constants.IP_ANY[netaddr.IPAddress(via).version]
            yield Route(dest=dest, properties=properties)


def list_ip_addresses(addresses, ip_addresses=None, port=None,
                      subnets=None):
    if port:
        # filter addresses by port IP addresses
        ip_addresses = set(ip_addresses) if ip_addresses else set()
        ip_addresses.update(list_port_ip_addresses(port=port,
                                                   subnets=subnets))
    if ip_addresses:
        addresses = [a for a in addresses if (hasattr(a, 'ip') and
                                              str(a.ip) in ip_addresses)]
    return addresses


def list_port_ip_addresses(port, subnets=None):
    fixed_ips = port['fixed_ips']
    if subnets:
        subnets = {subnet['id']: subnet for subnet in subnets}
        fixed_ips = [fixed_ip
                     for fixed_ip in fixed_ips
                     if fixed_ip['subnet_id'] in subnets]
    return [ip['ip_address'] for ip in port['fixed_ips']]


def get_port_device_name(addresses, port):
    for address in list_ip_addresses(addresses=addresses, port=port):
        return address.device.name

    msg = _("Port {!r} fixed IPs not found on server.".format(port['id']))
    raise ValueError(msg)


def get_vlan_device_name(addresses, ip_addresses):
    for address in list_ip_addresses(addresses=addresses,
            ip_addresses=ip_addresses):
        return address.device.name

    msg = _(
        "Fixed IPs {!r} not found on server.".format(' '.join(ip_addresses)))
    raise ValueError(msg)


def _get_ip_address_prefix_len_pairs(port, subnets):
    subnets = {subnet['id']: subnet for subnet in subnets}
    for fixed_ip in port['fixed_ips']:
        subnet = subnets.get(fixed_ip['subnet_id'])
        if subnet:
            yield (fixed_ip['ip_address'],
                   netaddr.IPNetwork(subnet['cidr']).prefixlen)


def arp_table(namespace=None):
    # 192.168.0.16  0x1  0x2  dc:a6:32:06:56:51  *  enp0s31f6
    regex_str = (r"([^ ]+)\s+(0x\d+)\s+(0x\d+)\s+(\w{2}\:\w{2}\:\w{2}\:\w{2}\:"
                 r"\w{2}\:\w{2})\s+([\w+\*]+)\s+([\-\w]+)")
    regex = re.compile(regex_str)
    arp_table = []
    cmd = ""
    if namespace:
        cmd = "sudo ip netns exec %s " % namespace
    cmd += "cat /proc/net/arp"
    arp_entries = shell.execute(cmd).stdout.split("\n")
    for line in arp_entries:
        m = regex.match(line)
        if m:
            arp_table.append(ARPregister(
                ip_address=m.group(1), hw_type=m.group(2),
                flags=m.group(3), mac_address=m.group(4),
                mask=m.group(5), device=m.group(6)))
    return arp_table


def list_iptables(version=constants.IP_VERSION_4, namespace=None):
    cmd = 'sudo '
    if namespace:
        cmd += 'ip netns exec %s ' % namespace
    cmd += ('iptables-save' if version == constants.IP_VERSION_4 else
            'ip6tables-save')
    return shell.execute(cmd).stdout


def list_listening_sockets(namespace=None):
    cmd = ''
    if namespace:
        cmd = 'sudo ip netns exec %s ' % namespace
    cmd += 'netstat -nlp'
    return shell.execute(cmd).stdout


class Route(HasProperties,
            collections.namedtuple('Route',
                                   ['dest', 'properties'])):

    @property
    def dest_ip(self):
        return netaddr.IPNetwork(self.dest)

    @property
    def via_ip(self):
        return netaddr.IPAddress(self.via)

    @property
    def src_ip(self):
        return netaddr.IPAddress(self.src)

    def __str__(self):
        properties_str = ' '.join('%s %s' % (k, v)
                                  for k, v in self.properties.items())
        return '%(dest)s %(properties)s' % {'dest': self.dest,
                                            'properties': properties_str}


class ARPregister(collections.namedtuple(
        'ARPregister',
        ['ip_address', 'hw_type', 'flags', 'mac_address', 'mask', 'device'])):

    def __str__(self):
        return '%s %s %s %s %s %s' % (self.ip_address, self.hw_type,
                                      self.flags, self.mac_address, self.mask,
                                      self.device)


def find_valid_cidr(valid_cidr='10.0.0.0/8', used_cidr=None):
    total_ips = netaddr.IPSet(netaddr.IPNetwork(valid_cidr))
    if used_cidr:
        used_network = netaddr.IPNetwork(used_cidr)
        netmask = used_network.netmask.netmask_bits()
        valid_ips = total_ips.difference(netaddr.IPSet(used_network))
    else:
        valid_ips = total_ips
        netmask = 24

    for ip in valid_ips:
        valid_network = netaddr.IPNetwork('%s/%s' % (ip, netmask))
        if valid_network in valid_ips:
            return valid_network.cidr

    exception_str = 'No valid CIDR found in %s' % valid_cidr
    if used_cidr:
        exception_str += ', used CIDR %s' % used_cidr
    raise Exception(exception_str)


def wait_for_interface_status(client, server_id, port_id, status,
                              ssh_client=None, mac_address=None):
    """Waits for an interface to reach a given status and checks VM NIC

    This method enhances the tempest one. Apart from checking the interface
    status returned by Nova, this methods access the VM to check if the NIC
    interface is already detected by the kernel.
    """
    body = waiters.wait_for_interface_status(client, server_id, port_id,
                                             status)

    if ssh_client and mac_address:
        ip_command = IPCommand(ssh_client)
        common_utils.wait_until_true(
            lambda: ip_command.get_nic_name_by_mac(mac_address),
            timeout=10,
            exception=RuntimeError('Interface with MAC %s not present in the '
                                   'VM' % mac_address))

    return body



# Copyright 2020 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class PortForwardingNegativeTestJSON(base.BaseNetworkTest):
    required_extensions = ['router', 'floating-ip-port-forwarding']

    @classmethod
    def resource_setup(cls):
        super().resource_setup()
        cls.ext_net_id = CONF.network.public_network_id

        # Create network, subnet, router and add interface
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router(data_utils.rand_name('router'),
                                       external_network_id=cls.ext_net_id)
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('63c0d406-99d5-11ea-bb37-0242ac130002')
    def test_mapping_same_fip_and_external_port_to_different_dest(self):
        port1 = self.create_port(self.network)
        port2 = self.create_port(self.network)
        fip_for_pf = self.create_floatingip()

        self.create_port_forwarding(
            fip_for_pf['id'],
            internal_port_id=port1['id'],
            internal_ip_address=port1['fixed_ips'][0]['ip_address'],
            internal_port=1111, external_port=2222,
            protocol="tcp")

        self.assertRaises(
            exceptions.BadRequest,
            self.create_port_forwarding,
            fip_for_pf['id'],
            internal_port_id=port2['id'],
            internal_ip_address=port2['fixed_ips'][0]['ip_address'],
            internal_port=3333, external_port=2222,
            protocol="tcp")

    @decorators.attr(type='negative')
    @decorators.idempotent_id('0c229a4c-9f28-11ea-bb37-0242ac130002')
    def test_mapping_different_external_ports_to_the_same_destination(self):
        port = self.create_port(self.network)
        fip_for_pf = self.create_floatingip()

        self.create_port_forwarding(
            fip_for_pf['id'],
            internal_port_id=port['id'],
            internal_ip_address=port['fixed_ips'][0]['ip_address'],
            internal_port=1111, external_port=3333,
            protocol="tcp")

        self.assertRaises(
            exceptions.BadRequest,
            self.create_port_forwarding,
            fip_for_pf['id'],
            internal_port_id=port['id'],
            internal_ip_address=port['fixed_ips'][0]['ip_address'],
            internal_port=1111, external_port=5555,
            protocol="tcp")

    @decorators.attr(type='negative')
    @decorators.idempotent_id('e9d3ffb6-e5bf-421d-acaa-ee6010dfbf14')
    def test_out_of_range_ports(self):
        port = self.create_port(self.network)
        fip_for_pf = self.create_floatingip()

        pf_params = {
            'internal_port_id': port['id'],
            'internal_ip_address': port['fixed_ips'][0]['ip_address'],
            'internal_port': 1111,
            'external_port': 3333,
            'protocol': "tcp"}
        pf = self.create_port_forwarding(fip_for_pf['id'], **pf_params)

        # Check: Invalid input for external_port update
        self.assertRaises(
            exceptions.BadRequest,
            self.update_port_forwarding,
            fip_for_pf['id'], pf['id'], external_port=108343)

        # Check: Invalid input for internal_port update
        self.assertRaises(
            exceptions.BadRequest,
            self.update_port_forwarding,
            fip_for_pf['id'], pf['id'], internal_port=108343)

        # Check: Invalid input for external_port create
        pf_params['internal_port'] = 4444
        pf_params['external_port'] = 333333
        self.assertRaises(
            exceptions.BadRequest,
            self.create_port_forwarding, fip_for_pf['id'], **pf_params)

        # Check: Invalid input for internal_port create
        pf_params['internal_port'] = 444444
        pf_params['external_port'] = 3333
        self.assertRaises(
            exceptions.BadRequest,
            self.create_port_forwarding, fip_for_pf['id'], **pf_params)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import exceptions

from neutron_tempest_plugin.common import utils


class NeutronTempestPluginException(exceptions.TempestException):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._properties = kwargs

    def __getattr__(self, name):
        try:
            return self._properties[name]
        except KeyError:
            pass

        msg = ("AttributeError: {!r} object has no attribute {!r}").format(
            self, name)
        raise AttributeError(msg)


class InvalidConfiguration(NeutronTempestPluginException):
    message = "Invalid Configuration"


class InvalidCredentials(NeutronTempestPluginException):
    message = "Invalid Credentials"


class InvalidServiceTag(NeutronTempestPluginException):
    message = "Invalid service tag"


class SSHScriptException(exceptions.TempestException):
    """Base class for SSH client execute_script() exceptions"""


class ShellError(NeutronTempestPluginException):
    pass


class ShellCommandFailed(ShellError):
    """Raised when shell command exited with non-zero status

    """
    message = ("Command %(command)r failed, exit status: %(exit_status)d, "
               "stderr:\n%(stderr)s\n"
               "stdout:\n%(stdout)s")


class SSHScriptFailed(ShellCommandFailed):
    message = ("Command %(command)r failed, exit status: %(exit_status)d, "
               "host: %(host)r\n"
               "script:\n%(script)s\n"
               "stderr:\n%(stderr)s\n"
               "stdout:\n%(stdout)s")


class ShellTimeoutExpired(ShellError):
    """Raised when shell command timeouts and has been killed before exiting

    """
    message = ("Command '%(command)s' timed out: %(timeout)d, "
               "stderr:\n%(stderr)s\n"
               "stdout:\n%(stdout)s")


class SSHScriptTimeoutExpired(ShellTimeoutExpired):
    message = ("Command '%(command)s', timed out: %(timeout)d "
               "host: %(host)r\n"
               "script:\n%(script)s\n"
               "stderr:\n%(stderr)s\n"
               "stdout:\n%(stdout)s")


# Patch SSHExecCommandFailed exception to make sure we can access to fields
# command, exit_status, STDOUT and STDERR when SSH client reports command
# failure
exceptions.SSHExecCommandFailed = utils.override_class(
    exceptions.SSHExecCommandFailed, ShellCommandFailed)

# Above code created a new SSHExecCommandFailed class based on top
# of ShellCommandFailed
assert issubclass(exceptions.SSHExecCommandFailed, ShellCommandFailed)
